Package: MsExperiment
Title: Infrastructure for Mass Spectrometry Experiments
Version: 1.14.0
Description: Infrastructure to store and manage all aspects related to
   a complete proteomics or metabolomics mass spectrometry (MS)
   experiment. The MsExperiment package provides light-weight and
   flexible containers for MS experiments building on the new MS
   infrastructure provided by the Spectra, QFeatures and related
   packages. Along with raw data representations, links to original
   data files and sample annotations, additional metadata or
   annotations can also be stored within the MsExperiment
   container. To guarantee maximum flexibility only minimal
   constraints are put on the type and content of the data within the
   containers.
Authors@R: c(person(given = "Laurent", family = "Gatto",
                    comment = c(ORCID = "0000-0002-1520-2268"),
                    email = "laurent.gatto@uclouvain.be",
                    role = c("aut", "cre")),
             person(given = "Johannes", family = "Rainer",
                    comment = c(ORCID = "0000-0002-6977-7147"),
                    email = "Johannes.Rainer@eurac.edu",
                    role = "aut"),
             person(given = "Sebastian", family = "Gibb",
                    comment = c(ORCID = "0000-0001-7406-4443"),
                    email = "mail@sebastiangibb.de",
                    role = "aut"),
             person(given = "Tuomas", family = "Borman",
                    comment = c(ORCID = "0000-0002-8563-8884"),
                    role = "ctb"))
Depends: R (>= 4.2), ProtGenerics (>= 1.35.2),
Imports: methods, S4Vectors, IRanges, Spectra, SummarizedExperiment,
        QFeatures, DBI, BiocGenerics
Suggests: testthat, knitr (>= 1.1.0), roxygen2, BiocStyle (>= 2.5.19),
        rmarkdown, rpx, mzR, MsDataHub, MsBackendSql (>= 1.3.2),
        RSQLite
License: Artistic-2.0
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/MsExperiment/issues
URL: https://github.com/RforMassSpectrometry/MsExperiment
biocViews: Infrastructure, Proteomics, MassSpectrometry, Metabolomics,
        ExperimentalDesign, DataImport
RoxygenNote: 7.3.3
Roxygen: list(markdown=TRUE)
Encoding: UTF-8
Collate: 'MsExperiment-db.R' 'MsExperiment-functions.R'
        'MsExperimentFiles.R' 'MsExperiment.R'
        'existMsExperimentFiles.R'
git_url: https://git.bioconductor.org/packages/MsExperiment
git_branch: RELEASE_3_23
git_last_commit: f47ab3c
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:26:29 UTC; biocbuild
Author: Laurent Gatto [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1520-2268>),
  Johannes Rainer [aut] (ORCID: <https://orcid.org/0000-0002-6977-7147>),
  Sebastian Gibb [aut] (ORCID: <https://orcid.org/0000-0001-7406-4443>),
  Tuomas Borman [ctb] (ORCID: <https://orcid.org/0000-0002-8563-8884>)
Maintainer: Laurent Gatto <laurent.gatto@uclouvain.be>

# Generated by roxygen2: do not edit by hand

export("experimentFiles<-")
export("otherData<-")
export("qdata<-")
export("sampleData<-")
export(MsExperiment)
export(MsExperimentFiles)
export(dbWriteSampleData)
export(existMsExperimentFiles)
export(experimentFiles)
export(linkSampleData)
export(otherData)
export(qdata)
export(readMsExperiment)
export(sampleData)
export(spectraSampleIndex)
exportClasses(MsExperiment)
exportClasses(MsExperimentFiles)
exportMethods("[")
exportMethods("spectra<-")
exportMethods(filterSpectra)
exportMethods(length)
exportMethods(show)
exportMethods(spectra)
importClassesFrom(IRanges,SimpleCharacterList)
importClassesFrom(QFeatures,QFeatures)
importClassesFrom(S4Vectors,Annotated)
importClassesFrom(S4Vectors,List)
importClassesFrom(S4Vectors,SimpleList)
importClassesFrom(Spectra,Spectra)
importClassesFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(DBI,dbDisconnect)
importFrom(DBI,dbExistsTable)
importFrom(DBI,dbGetQuery)
importFrom(DBI,dbListTables)
importFrom(DBI,dbRemoveTable)
importFrom(DBI,dbWriteTable)
importFrom(IRanges,CharacterList)
importFrom(IRanges,LogicalList)
importFrom(ProtGenerics,spectra)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,List)
importFrom(S4Vectors,findMatches)
importFrom(S4Vectors,from)
importFrom(S4Vectors,to)
importFrom(Spectra,msLevel)
importFrom(methods,"slot<-")
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,new)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(methods,validObject)
importMethodsFrom(BiocGenerics,dbconn)
importMethodsFrom(ProtGenerics,"spectra<-")
importMethodsFrom(ProtGenerics,filterSpectra)
importMethodsFrom(S4Vectors,"mcols<-")
importMethodsFrom(S4Vectors,"metadata<-")
importMethodsFrom(S4Vectors,mcols)
importMethodsFrom(S4Vectors,metadata)
importMethodsFrom(Spectra,Spectra)
importMethodsFrom(Spectra,peaksVariables)
importMethodsFrom(Spectra,selectSpectraVariables)
importMethodsFrom(Spectra,spectraVariables)
importMethodsFrom(methods,show)

# Managing mass spectrometry experiments

[![Project Status: Active - The project has reached a stable, usable state and is being actively developed.](http://www.repostatus.org/badges/latest/active.svg)](http://www.repostatus.org/#active)
[![R-CMD-check-bioc](https://github.com/RforMassSpectrometry/MsExperiment/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/RforMassSpectrometry/MsExperiment/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov](https://codecov.io/gh/rformassspectrometry/MsExperiment/branch/main/graph/badge.svg?token=IEBQZJAJVI)](https://codecov.io/gh/rformassspectrometry/MsExperiment)
[![license](https://img.shields.io/badge/license-Artistic--2.0-brightgreen.svg)](https://opensource.org/licenses/Artistic-2.0)
[![years in bioc](http://bioconductor.org/shields/years-in-bioc/MsExperiment.svg)](https://bioconductor.org/packages/release/bioc/html/MsExperiment.html)
[![Ranking by downloads](http://bioconductor.org/shields/downloads/release/MsExperiment.svg)](https://bioconductor.org/packages/stats/bioc/MsExperiment/)
[![build release](http://bioconductor.org/shields/build/release/bioc/MsExperiment.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/MsExperiment/)
[![build devel](http://bioconductor.org/shields/build/devel/bioc/MsExperiment.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/MsExperiment/)


The `MsExperiment` package provides the `MsExperiment` class that can
be used to store and manage all aspects related to a complete
proteomics or metabolomics mass spectrometry experiment. This includes

- experimental design
- data files
- raw data: spectra and chromatograms
- proteomics identification results and protein database
- quantitative features

The respective data handling functionality is handled by dedicated
packages such as `Spectra` for spectra data, `QFeatures` for
quantitative features, `Biostrings` for fasta files, ...

See the package [homepage](https://rformassspectrometry.github.io/MsExperiment)
for more information.

This package is part of the RforMassSpectrometry initiative:
https://www.rformassspectrometry.org/


# Installation

The package can be installed with

```r
install.packages("BiocManager")
BiocManager::install("MsExperiment")
```

# Contributions

Contributions are highly welcome and should follow the [contribution
guidelines](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html#contributions).
Also, please check the coding style guidelines in the [RforMassSpectrometry
vignette](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html).

library("testthat")
library("MsExperiment")
library("S4Vectors")
library("Spectra")
library("MsDataHub")

fls <- unname(c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
                MsDataHub::X20171016_POOL_POS_3_105.134.mzML()))

mse <- MsExperiment()
df <- data.frame(sample = c("QC1", "QC2"),
                 time = c(1, 2),
                 mzML_file = basename(fls))
sampleData(mse) <- DataFrame(df)
experimentFiles(mse) <- MsExperimentFiles(mzML_file = fls,
                                          other_file = "other_file.txt")
spectra(mse) <- Spectra(fls)

test_check("MsExperiment")

#' @param object The `existMsExperimentFiles()` function works with
#'     either an instance of `MsExperimentFiles` or `MsExperiment`.
#'
#' @export
#'
#' @importFrom IRanges LogicalList
#'
#' @rdname MsExperimentFiles
existMsExperimentFiles <- function(object) {
    if (inherits(object, "MsExperiment"))
        object <- experimentFiles(object)
    stopifnot(inherits(object, "MsExperimentFiles"))
    res <- LogicalList(lapply(object, file.exists))
    message(paste0(names(res), ": ", vapply(res, sum, integer(1)),
                   " out of ", lengths(res), " exist(s)\n"))
    invisible(res)
}

test_that("existMsExperimentFiles works", {
    res <- existMsExperimentFiles(MsExperimentFiles())
    expect_true(length(res) == 0)

    x <- MsExperimentFiles(a = "some_file", b = "other_file", d = tempfile())
    res <- existMsExperimentFiles(x)
    expect_true(length(res) == length(x))
    expect_true(all(all(!res)))

    expect_message(res <- existMsExperimentFiles(mse), "mzML_file: 2 out")
    expect_true(all(res[[1L]]))
    expect_false(res[[2L]])
})

#' @include MsExperimentFiles.R

#' @title Managing Mass Spectrometry Experiments
#'
#' @aliases MsExperiment-class MsExperiment
#'
#' @description
#'
#' The `MsExperiment` class allows the storage and management of all
#' aspects related to a complete proteomics or metabolomics mass
#' spectrometry experiment. This includes experimantal design (i.e. a table
#' with samples), raw mass spectromtry data as spectra and chromatograms,
#' quantitative features, and identification data or any other relevant data
#' files.
#'
#' For details, see https://rformassspectrometry.github.io/MsExperiment
#'
#' This package is part of the RforMassSpectrometry initiative:
#' https://www.rformassspectrometry.org/
#'
#' @section General information:
#'
#' An experiment is typically composed of several items
#'
#' - Description and information (covariates etc) of each sample from
#'   the experiment. These are stored in the `sampleData` slot as a
#'   `DataFrame`, each row describing a sample with columns containing
#'   all relevant information on that sample.
#'
#' - Files to data or annotations. These are stored in the
#'   `@experimentFiles` slot as an instance of class `MsExperimentFiles`.
#'
#' - General metadata about the experiment, stored as a `list` in the
#'   `@metadata` slot.
#'
#' - Mass spectrometry data. Sectra and their metadata are stored as
#'   an `[Spectra::Spectra()]` object in the `spectra` slot. Chromatographic
#'   data is not yet supported but will be stored as a `Chromatograms()`
#'   object in the `@chromatorgrams` slot.
#'
#' - Quantification data is stored as `QFeatures` or
#'   `SummarizedExperiment` objects in the `@qdata` slot and can be accessed or
#'   replaced with the `qdata()` or `qdata<-` functions, respectively.
#'
#' - Any additional data, be it other spectra data, or proteomics
#'   identification data (i.e peptide-spectrum matches defined as
#'   `PSM` objects) can be added as elements to the list stored in
#'   the `otherData` slot.
#'
#' The *length* of a `MsExperiment` is defined by the number of samples (i.e.
#' the number of rows of the object's `sampleData`). A `MsExperiment` with two
#' samples will thus have a length of two, independently of the number of files
#' or length of raw data in the object. This also defines the subsetting of the
#' object using the `[` function which will always subset by samples. See the
#' section for filtering and subsetting below for more information.
#'
#' `MsExperiment` objects can be created using the `MsExperiment()` function
#' providing the data with the parameters listed below. If the
#' [Spectra::Spectra()]
#' object provided with the `spectra` param uses a `MsBackendSql` backend,
#' sample data could be retrieved from the associated SQL database (see
#' section *Using `MsExperiment` with `MsBackendSql`* in the vignette for
#' details). Alternatively, it is also possible to subsequently add data and
#' information to an existing `MsExperiment`.
#' Finally, with the [readMsExperiment()] function it is possible to create
#' a `MsExperiment` by importing MS spectra data directly from provided data
#' files. See examples below or the package vignette for more information.
#'
#' @section Accessing data:
#'
#' Data from an `MsExperiment` object can be accessed with the dedicated
#' accessor functions:
#'
#' - `experimentFiles()`, `experimentFiles<-`: gets or sets experiment files.
#'
#' - `length()`: get the *length* of the object which represents the number of
#'   samples availble in the object's `sampleData`.
#'
#' - `metadata()`, `metadata<-`: gets or sets the object's metadata.
#'
#' - `sampleData()`, `sampleData<-`: gets or sets the object's sample data
#'   (i.e. a `DataFrame` containing sample descriptions).
#'
#' - `spectra()`, `spectra<-`: gets or sets spectra data. `spectra()` returns a
#'   [Spectra::Spectra()] object, `spectra<-` takes a `Spectra` data as input
#'   and returns the updated `MsExperiment`.
#'
#' - `spectraSampleIndex()`: depending on parameter `duplicates` it returns
#'   either an `integer` (`duplicates = "first"`, the default) or a `list`
#'   (`duplicates = "keep"`) of length equal to the number of spectra within
#'   the object with the indices of the sample(s) (in `sampleData()`) a
#'   spectrum is assigned to. With `duplicates = "first"`, an `integer` with
#'   the index is returned for each spectrum. If a spectrum was assigned to
#'   more than one sample a warning is shown and only the first sample index
#'   is returned for that spectrum. For `duplicates = "keep"`, assignments are
#'   returned as a `list` of `integer` vectors, each element being the
#'   index(es) of the sample(s) a spectrum is assigned to. For spectra that are
#'   not linked to any sample an `NA_integer_` is returned as index for
#'   `duplicates = "first"` and an empty integer (`integer()`) for
#'   `duplicates = "keep"`.
#'   Note that the default `duplicates = "first"` will work in almost all use
#'   cases, as generally, a spectrum will be assigned to a single sample.
#'
#' - `qdata()`, `qdata<-`: gets or sets the quantification data, which can be a
#'   `QFeatures` or `SummarizedExperiment`.
#'
#' - `otherData()` , `otherData<-`: gets or sets the addition data
#'   types, stored as a `List` in the object's `otherData` slot.
#'
#' @section Linking sample data to other experimental data:
#'
#' To start with, an `MsExperiment` is just a loose collection of files and data
#' related to an experiment, no explicit links or associactions are present
#' between the samples and related data. Such links can however be created with
#' the `linkSampleData()` function. This function can establish links between
#' individual (or all) samples within the object's `sampleData` to individual,
#' or multiple, data elements or files, such as `Spectra` or raw data files.
#'
#' The presence of such links enables a (consistent) subsetting of an
#' `MsExperiment` by samples. Thus, once the link is defined, any subsetting by
#' sample will also correctly subset the linked data. All other, not linked,
#' data elements are always retained as in the original `MsExperiment`.
#'
#' To be able to link different elements within an `MsExperiment` it is also
#' required to *identify* them with a consistent naming scheme. The naming
#' scheme of slots and data elements within follows an SQL-like scheme, in which
#' the variable (element) is identified by the name of the database table,
#' followed by a `"."` and the name of the database table column. For
#' `MsExperiment`, the naming scheme is defined as
#' `"<slot name>.<element name>"`. A column called `"sample_name"` within the
#' `sampleData` data frame can thus be addressed with
#' `"sampleData.sample_name"`, while `spectra.msLevel` would represent the
#' spectra variable called `msLevel` within the `Spectra` stored in the
#' `spectra` slot.
#'
#' Links between sample data rows and any other data element are stored as
#' `integer` matrices within the `@sampleDataLinks` slot of the object (see also
#' the vignette for examples and illustrations). The first column of a matrix
#' is always the index of the sample, and the second column the index of the
#' element that is linked to that sample, with one row per element.
#' Links can be defined/added with the `linkSampleData()` function which adds
#' a relationship between rows in `sampleData` to elements in any other data
#' within the `MsExperiment` that are specified with parameter `with`.
#' `linkSampleData()` supports two different ways to define the link:
#'
#' - Parameter `with` defines the data to which the link should be established.
#'   To link samples to raw data files that would for example be available as a
#'   `character` in an element called `"raw_files"` within the object's
#'   `experimentFiles`, `with = experimentFiles.raw_files` would have to be
#'   used. Next it is required to specify which samples should be linked with
#'   which elements in `with`. This needs to be defined with the parameters
#'   `sampleIndex` and `withIndex`, both are expected to be `integer` vectors
#'   specifying which sample in `sampleData` should be linked to which element
#'   in `with` (see examples below or vignette for examples and details).
#'
#' - As an alternative way, a link could be defined with an SQL-like syntax
#'   that relates a column in `sampleData` to a column/element in the data to
#'   which the link should be established. To link for example individual
#'   spectra to the corresponding samples
#'   `with = "sampleData.raw_file = spectra.dataOrigin"` could be used assuming
#'   that `sampleData` contains a column named `"raw_file"` with the (full path)
#'   of the raw data file for each sample from which the spectra were imported.
#'   In this case both `sampleIndex` and `withIndex` can be omitted, but it is
#'   expected/required that the columns/elements from `sampleData` and the data
#'   element to which the link should be established contain matching values.
#'
#' Note that `linkSampleData` will **replace** a previously existing link to the
#' same data element.
#'
#' - `spectraSampleIndex()` is a convenience function that extracts for each
#'   spectrum in the object's `spectra()` the index of the sample it is
#'   associated with (see function's help above for more information).
#'
#' @section Subsetting and filtering:
#'
#' - `[`: `MsExperiment` objects can be subset **by samples** with `[i]`
#'   where `i` is the index or a logical defining to which samples the data
#'   should be subset. Subsetting by sample will (correctly) subset all
#'   linked data to the respective samples. If multiple samples are linked to
#'   the same data element, subsetting might duplicate that data element. This
#'   duplication of *n:m* relationships between samples to elements does however
#'   not affect data consistency (see examples below for more information).
#'   Not linked data (slots) will be returned as they are. Subsetting in
#'   arbitrary order is supported.
#'   See the vignette for details and examples.
#'
#' - `filterSpectra()`: subsets the `Spectra` within an `MsExperiment` using a
#'   provided filter function (parameter `filter`). Parameters for the filter
#'   function can be passed with parameter `...`. Any of the filter functions
#'   of a [Spectra::Spectra()] object can be passed with parameter `filter`.
#'   Possibly present relationships between samples and spectra (*links*, see
#'   also `linkSampleData()`) are updated. Filtering affects only the spectra
#'   data of the object, none of the other slots and data (e.g. `sampleData`)
#'   are modified.
#'   The function returns an `MsExperiment` with the filtered `Spectra` object.
#'
#' @return See help of the individual functions.
#'
#' @param spectra [Spectra::Spectra()] object with the MS spectra data of the
#'     experiment.
#'
#' @param drop for `[`: ignored.
#'
#' @param duplicates for `spectraSampleIndex()`: `character(1)` defining the
#'     type of result returned by `spectraSampleIndex()`. With
#'     `duplicates = "first"` an `integer` vector is returned with the first
#'     match while `duplicates = "keep"` returns a `list` of `integer` with
#'     the index of all matches.
#'
#' @param experimentFiles [MsExperimentFiles()] defining (external) files
#'     to data or annotation.
#'
#' @param filter for `filterSpectra()`: any filter function supported by
#'     [Spectra::Spectra()] to filter the spectra object (such as `filterRt` or
#'     `filterMsLevel`). Parameters for the filter function can be passed
#'     through `...`.
#'
#' @param i for `[`: an `integer`, `character` or `logical` referring to the
#'     indices or names (rowname of `sampleData`) of the samples to subset.
#'
#' @param j for `[`: not supported.
#'
#' @param object an `MsExperiment`.
#'
#' @param otherData `List` with arbitrary additional (*other*) information or
#'     data.
#'
#' @param qdata `QFeatures` or `SummarizedExperiment` with the quantification
#'     data.
#'
#' @param sampleData `DataFrame` (or `data.frame`) with information on
#'     individual samples of the experiment.
#'
#' @param sampleIndex for `linkSampleData()`: `integer` with the indices of the
#'     samples in `sampleData(object)` that should be linked.
#'
#' @param subsetBy for `linkSampleData()`: optional `integer(1)` defining the
#'     dimension on which the subsetting will occurr on the linked data.
#'     Defaults to `subsetBy = 1L` thus subsetting will happen on the first
#'     dimension (rows or elements).
#'
#' @param with for `linkSampleData()`: `character(1)` defining the data to which
#'     samples should be linked. See section *Linking sample data to other
#'     experimental data* for details.
#'
#' @param withIndex for `linkSampleData()`: `integer` with the indices of the
#'     elements in `with` to which the samples (specified by `sampleIndex`)
#'     should be linked to.
#'
#' @param x an `MsExperiment`.
#'
#' @param ... optional additional parameters. For `filterSpectra()`: parameters
#'     to be passed to the filter function (parameter `filter`).
#'
#' @name MsExperiment
#'
#' @importFrom S4Vectors List DataFrame
#'
#' @importClassesFrom S4Vectors List
#'
#' @author Laurent Gatto, Johannes Rainer
#'
#' @examples
#'
#' ## An empty MsExperiment object
#' msexp <- MsExperiment()
#' msexp
#'
#' example(MsExperimentFiles)
#' experimentFiles(msexp) <- fls
#' msexp
#'
#' ## Linking samples to data elements
#'
#' ## Create a small experiment
#' library(S4Vectors)
#' mse <- MsExperiment()
#' sd <- DataFrame(sample_id = c("QC1", "QC2"),
#'                 sample_name = c("QC Pool", "QC Pool"),
#'                 injection_idx = c(1, 3))
#' sampleData(mse) <- sd
#'
#' ## define file names containing spectra data for the samples and
#' ## add them, along with other arbitrary files to the experiment
#' fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
#'          MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
#' experimentFiles(mse) <- MsExperimentFiles(
#'     mzML_files = fls,
#'     annotations = "internal_standards.txt")
#'
#' ## Link samples to data files: first sample to first file in "mzML_files",
#' ## second sample to second file in "mzML_files"
#' mse <- linkSampleData(mse, with = "experimentFiles.mzML_files",
#'     sampleIndex = c(1, 2), withIndex = c(1, 2))
#'
#' ## Link all samples to the one file in "annotations"
#' mse <- linkSampleData(mse, with = "experimentFiles.annotations",
#'     sampleIndex = c(1, 2), withIndex = c(1, 1))
#' mse
#'
#' ## Import the spectra data and add it to the experiment
#' library(Spectra)
#' spectra(mse) <- Spectra(fls, backend = MsBackendMzR())
#'
#' ## Link each spectrum to the respective sample. We use the alternative
#' ## link definition that does not require sampleIndex and withIndex but
#' ## links elements based on matching values in the specified data elements.
#' ## We need to add the full file name as an additional column to sampleData
#' ## in order to allow matching this file names with the value in
#' ## spectra(mse)$dataOrigin which contains the original file names from which
#' ## the spectra were imported.
#' sampleData(mse)$raw_file <- normalizePath(fls)
#'
#' ## The links can be added using the short notation below
#' mse <- linkSampleData(mse, with = "sampleData.raw_file = spectra.dataOrigin")
#' mse
#'
#' ## With sampleData links present, any subsetting of the experiment by sample
#' ## will ensure that all linked elements are subset accordingly
#' b <- mse[2]
#' b
#' sampleData(b)
#' experimentFiles(b)$mzML_files
#'
#' ## The `spectraSampleIndex()` function returns, for each spectrum, the
#' ## index in the object's `sampleData` to which it is linked/assigned
#' spectraSampleIndex(mse)
#'
#' ## Subsetting with duplication of n:m sample to data relationships
#' ##
#' ## Both samples were assigned above to one "annotation" file in
#' ## `experimentFiles`:
#' experimentFiles(mse[1])[["annotations"]]
#' experimentFiles(mse[2])[["annotations"]]
#'
#' ## Subsetting will always keep the relationship between samples and linked
#' ## data elements. Subsetting will however possibly duplicate data elements
#' ## that are shared among samples. Thus, while in the original object the
#' ## element "annotations" has a single entry, subsetting with [1:2] will
#' ## result in an MsExperiment with duplicated entries in "annotations"
#' experimentFiles(mse)[["annotations"]]
#' experimentFiles(mse[1:2])[["annotations"]]
#'
#' ## Spectra within an MsExperiment can be filtered/subset with the
#' ## `filterSpectra` function and any of the filter functions supported
#' ## by `Spectra` objects. Below we restrict the spectra data to spectra
#' ## with a retention time between 200 and 210 seconds.
#' res <- filterSpectra(mse, filterRt, rt = c(200, 210))
#' res
#'
#' ## The object contains now much less spectra. The retention times for these
#' rtime(spectra(res))
#'
#' ## Relationship between samples and spectra was preserved by the filtering
#' a <- res[1L]
#' spectra(a)
NULL

#' @name MsExperiment-class
#'
#' @docType class
#'
#' @exportClass MsExperiment
#'
#' @slot experimentFiles An instance of class `MsExperimentFiles` or `NULL`.
#'
#' @slot spectra An instance of class `Spectra` or `NULL`.
#'
#' @slot qdata An instance of class `QFeatures`, `SummarizedExperiment` or
#'     `NULL`.
#'
#' @slot otherData A `List` to store any additional data objects.
#'
#' @slot sampleData A `DataFrame` documenting the experimental design.
#'
#' @slot sampleDataLinks A `List` with link definitions between samples and
#'     data elements. Should not be directly accessed or modified by the user.
#'
#' @slot metadata A `list` to store additional metadata.
#'
#' @rdname MsExperiment
#'
#' @importClassesFrom S4Vectors SimpleList
#'
#' @importClassesFrom S4Vectors Annotated
#'
#' @importFrom S4Vectors DataFrame
setClass("MsExperiment",
         contains = "Annotated",
         slots = c(
             experimentFiles = "MsExperimentFiles",
             spectra = "Spectra_OR_Null",
             qdata = "QFeatures_OR_SummarizedExperiment_OR_Null",
             ## chromatograms = "Chromatograms",
             otherData = "List",
             sampleData = "DataFrame",
             sampleDataLinks = "List"),
         prototype = prototype(
             experimentFiles = MsExperimentFiles(),
             spectra = NULL,
             qdata = NULL,
             otherData = List(),
             sampleData = DataFrame(),

             sampleDataLinks = new(
                 "SimpleList", elementMetadata =
                                   DataFrame(subsetBy = integer())))
         )

#' @rdname MsExperiment
#'
#' @export
MsExperiment <- function(experimentFiles = MsExperimentFiles(),
                         otherData = List(),
                         qdata = NULL,
                         sampleData = DataFrame(),
                         spectra = NULL) {
    if (!length(sampleData) && length(spectra) &&
        inherits(spectra, "Spectra") &&
        inherits(spectra@backend, "MsBackendSql")) {
        sampleData <- .db_get_sample_data(spectra)
        slink <- .db_get_sample_spectra_link(spectra)
    } else slink <- matrix(ncol = 2, nrow = 0)
    res <- new("MsExperiment", experimentFiles = experimentFiles,
               otherData = otherData, qdata = qdata,
               sampleData = DataFrame(sampleData), spectra = spectra)
    if (length(slink)) {
        res@sampleDataLinks[["spectra"]] <- slink
        mcols(res@sampleDataLinks)["spectra", "subsetBy"] <- 1L
    }
    res
}

#' @rdname MsExperiment
#'
#' @param object An instance of class `MsExperiment`.
#'
#' @importFrom Spectra msLevel
#'
#' @exportMethod show
setMethod("show", "MsExperiment", function(object) {
    mess <- "Object of class"
    if (.ms_experiment_is_empty(object))
        mess <- "Empty object of class"
    cat(mess, class(object), "\n")
    ## Show experiment files
    if (length(experimentFiles(object)))
        cat(" Files:", paste(names(experimentFiles(object)),
                             collapse = ", "), "\n")
    ## Show spectra
    if (!is.null(object@spectra)) {
        mstab <- table(msLevel(object@spectra))
        cat(" Spectra:", paste0("MS", names(mstab), " (", mstab, ")"),
            "\n")
    }
    ## Show quantitative data
    if (!is.null(object@qdata)) {
        if (inherits(object@qdata, "SummarizedExperiment"))
            cat( " SummarizedExperiment:",
                nrow(object@qdata), "feature(s)\n")
        else ## QFeatures
            cat( " QFeatures:",
                length(object@qdata), "assay(s)\n")

    }
    ## Show other data
    if (length(object@otherData)) {
        cat(" Other data:",
            paste(names(object@otherData), collapse = ", "),
            "\n")
    }
    ## Show experiment/sample data
    if (nrow(object@sampleData)) {
        cat(" Experiment data:",
            nrow(object@sampleData), "sample(s)\n")
    }
    lnks <- object@sampleDataLinks
    if (length(lnks)) {
        cat(" Sample data links:\n")
        for (i in seq_along(lnks)) {
            if (mcols(lnks)$subsetBy[i] == 2)
                cols <- " column(s).\n"
            else cols <- " element(s).\n"
            cat("  - ", names(lnks)[i], ": ", length(unique(lnks[[i]][, 1L])),
                " sample(s) to ", length(unique(lnks[[i]][, 2L])), cols,
                sep = "")
        }
    }
})

#' @rdname MsExperiment
#'
#' @exportMethod length
setMethod("length", "MsExperiment", function(x) {
    nrow(sampleData(x))
})

## ------------------------------##
##     Getters and setters       ##
## ------------------------------##

#' @export
#'
#' @importFrom ProtGenerics spectra
#'
#' @rdname MsExperiment
setMethod("spectra", "MsExperiment", function(object) object@spectra)

#' @importMethodsFrom ProtGenerics spectra<-
#'
#' @export
#'
#' @rdname MsExperiment
setReplaceMethod("spectra", "MsExperiment", function(object, value) {
    stopifnot(inherits(value, "Spectra"))
    stopifnot(inherits(object, "MsExperiment"))
    object@spectra <- value
    object
})


#' @rdname MsExperiment
#'
#' @export
otherData <- function(object) {
    stopifnot(inherits(object, "MsExperiment"))
    object@otherData

}

#' @rdname MsExperiment
#'
#' @export
`otherData<-` <- function(object, value) {
    object@otherData <- value
    object
}

#' @rdname MsExperiment
#'
#' @export
linkSampleData <- function(object, with = character(),
                           sampleIndex = seq_len(nrow(sampleData(object))),
                           withIndex = integer(), subsetBy = 1L) {
    if (!length(with))
        return(object)
    subsetBy <- as.integer(subsetBy[1L])
    if (is.na(subsetBy))
        stop("'subsetBy' needs to be an integer of length 1")
    if (!length(withIndex)) {
        link_string <- .parse_join_string(with)
        if (link_string[1L] == "sampleData") {
            from <- paste0(link_string[c(1, 2)], collapse = ".")
            to_slot <- link_string[3L]
            with <- paste0(link_string[c(3, 4)], collapse = ".")
        } else if (link_string[3L] == "sampleData") {
            from <- paste0(link_string[c(3, 4)], collapse = ".")
            to_slot <- link_string[1L]
            with <- paste0(link_string[c(1, 2)], collapse = ".")
        } else stop("one of the slot names has to be 'sampleData'.")
        link <- .link_matrix(.get_element(object, from),
                             .get_element(object, with))
        if (nrow(link) == 0)
            warning("no matches found for '", with, "'")
    } else {
        sampleIndex <- as.integer(sampleIndex)
        withIndex <- as.integer(withIndex)
        sampleIndex <- sampleIndex[!is.na(sampleIndex)]
        withIndex <- withIndex[!is.na(withIndex)]
        if (length(sampleIndex) != length(withIndex))
            stop("Length of 'sampleIndex' and 'withIndex' have to match")
        link <- cbind(sampleIndex, withIndex)
    }
    withl <- unlist(strsplit(with, split = ".", fixed = TRUE))
    if (withl[1L] %in% c("spectra", "qdata")) {
        with <- withl[1L]
        if (with == "qdata")
            subsetBy <- 2L
    } else if (length(withl) < 2)
        stop("'with' should be a 'character' with the name of the slot and",
             " the name of element separated by a '.'. ",
             "See ?linkSampleData for examples")
    .add_sample_data_link(object, link, with = with, subsetBy = subsetBy)
}

#' @rdname MsExperiment
#'
#' @export
setMethod("[", "MsExperiment", function(x, i, j, ..., drop = FALSE) {
    if (!missing(j))
        stop("Only subsetting with '[i]' is supported.")
    li <- length(i)
    if (is.character(i)) {
        i <- match(i, rownames(sampleData(x)))
        if (any(is.na(i)))
            warning(sum(is.na(i)), " of ", li, " values could not be ",
                    "matched to rownames of 'sampleData(x)'")
        i <- i[!is.na(i)]
    }
    if (is.logical(i)) {
        if (li != nrow(sampleData(x)))
            stop("if 'i' is logical its length has to match the number of ",
                 "samples in 'x'.")
        i <- which(i)
    }
    if (any(i < 0)) {
        if (all(i < 0))
            i <- seq_along(x)[i]
        else stop("Mixing positive and negative indices is not supported.")
    }
    .extractSamples(x, i, newx = x)
})

#' @rdname MsExperiment
#'
#' @importMethodsFrom Spectra selectSpectraVariables
#'
#' @importMethodsFrom Spectra spectraVariables
#'
#' @importMethodsFrom Spectra peaksVariables
#'
#' @importMethodsFrom ProtGenerics filterSpectra
#'
#' @export
setMethod(
    "filterSpectra", c("MsExperiment", "function"),
    function(object, filter, ...) {
        ls <- length(spectra(object))
        if (!ls)
            return(object)
        have_links <- length(.sample_data_links(object, "spectra")) > 0
        if (have_links)
            object@spectra$._SPECTRA_IDX <- seq_len(ls)
        object@spectra <- filter(object@spectra, ...)
        if (have_links) {
            if (ls != length(spectra(object)))
                object <- .update_sample_data_links_spectra(object)
            svs <- unique(c(spectraVariables(spectra(object)),
                            peaksVariables(spectra(object))))
            object@spectra <- selectSpectraVariables(
                object@spectra, svs[svs != "._SPECTRA_IDX"])
        }
        object
    })

.sample_data_links <- function(x, name = character()) {
    if (length(name))
        x@sampleDataLinks[names(x@sampleDataLinks) %in% name]
    else x@sampleDataLinks
}

#' Check that a *link* matrix is in the correct format:
#'
#' - matrix
#' - integer
#' - two columns
#' - values within 1:length
#'
#' @author Johannes Rainer
#'
#' @noRd
.valid_link <- function(x, nfrom = 0L, nto = 0L) {
    if (!is.matrix(x) || !is.integer(x[1, 1]))
        stop("'x' needs to be an 'integer' 'matrix'")
    if (ncol(x) != 2)
        stop("A link matrix is expected to have 2 columns")
    if (any(x[, 1] > nfrom))
        stop("indices in the first column need to be <= ", nfrom)
    if (any(x[, 2] > nto))
        stop("indices in the second column need to be <= ", nto)
    if (any(x < 1))
        stop("indices can not be smaller than 1")
}

#' The validity of the link matrix is evaluated only when adding the link. Also,
#' possibly existing links between the same entities will be **overwritten**.
#'
#' @param x `LinkedMsExperiment`.
#'
#' @param link two-column `integer` `matrix` with the link.
#'
#' @param with `character(1)` with the definition of the element to which the
#'     link was established (e.g. `"spectra.dataOrigin"` or
#'     `experimentFiles.raw_files`. See also `.get_element` for details.
#'
#' @param subsetBy `integer(1)` defining on which dimension (for `with` with
#'     dimensions > 0) the subsetting should be done.
#'
#' @author Johannes Rainer
#'
#' @importMethodsFrom S4Vectors mcols<- mcols
#'
#' @noRd
.add_sample_data_link <- function(x, link = matrix(), with, subsetBy = 1L) {
    nfrom <- nrow(sampleData(x))
    if (nrow(link) == 0 || nfrom == 0)
        return(x)
    nto <- .nelements(.get_element(x, with), subsetBy) # support link to column
    if (nto == 0)
        stop("'", with, "' is empty. Can not link to empty data")
    .valid_link(link, nfrom, nto)
    if (any(names(.sample_data_links(x)) == with))
        warning("Overwriting previously present link '", with, "'")
    x@sampleDataLinks[[with]] <- unname(link)
    mcols(x@sampleDataLinks)[with, "subsetBy"] <- subsetBy
    x
}

#' Helper function to return an element (slot or
#'
#' The convention to name an element in a slot is `<slot name>.<field name>`. If
#' field name is missing the slot is returned.
#' Note that this fails if the object stored in slot is not a `matrix` or does
#' not have the `$` method implemented. It will work for `data.frame`,
#' `DataFrame`, `list`, `Spectra`.
#'
#' @param x any `S4Object` with slots.
#'
#' @param name `character(1)` defining the element name.
#'
#' @return the requested element which can be a slot or a column in slot.
#'
#' @author Johannes Rainer
#'
#' @importFrom methods slotNames slot slot<-
#'
#' @noRd
.get_element <- function(x, name = "sampleData") {
    name <- unlist(strsplit(name[1L], split = ".", fixed = TRUE))
    slt <- name[1L]
    if (!any(slotNames(x) == slt))
        stop("No slot named '", slt, "' available in 'x'")
    res <- slot(x, slt)
    if (length(name) > 1) {
        el <- paste0(name[-1L], collapse = ".")
        if (is.matrix(res))
            res <- res[, el]
        else
            res <- do.call("$", args = list(res, el))
    }
    res
}

#' Setting/replacing an element within a slot of an `LinkedMsExperiment`. This
#' function does **not** perform any checks except for the presence of a slot.
#'
#' @importFrom methods slot slot<-
#'
#' @author Johannes Rainer
#'
#' @noRd
#'
#' @examples
#'
#' ## Add a new spectra variable to a Spectra
#' .set_element(mse, "spectra.new.value", 1:3)
#'
#' ## Add a new metadata element
#' .set_element(mse, "metadata.new_entry", data.frame(1:3))
.set_element <- function(x, name = character(), value = NULL) {
    if (length(name) == 0)
        return(x)
    name <- unlist(strsplit(name[1L], split = ".", fixed = TRUE))
    slt <- name[1L]
    if (!any(slotNames(x) == slt))
        stop("No slot named '", slt, "' available in 'x'")
    if (length(name) == 1)
        slot(x, slt, check = FALSE) <- value
    else {
        el <- paste0(name[-1L], collapse = ".")
        if (length(dim(slot(x, slt))))
            slot(x, slt)[, el] <- value
        else
            slot(x, slt) <- do.call("$<-", list(slot(x, slt), el, value))
    }
    x
}

#' @param x `character` similar to a join statement, e.g.
#'     `"sampleData.mzML = spectra.dataOrigin"`
#'
#' @return `character` with slot from, column from, slot to, column to.
#'
#' @author Johannes Rainer
#'
#' @noRd
#'
#' @examples
#'
#' x <- "sampleData.mzml_file = spectra.dataOrigin"
#' .parse_join_string(x)
#'
#' .parse_join_string("sampleData.mzml.file= spectra.dataOrigin")
.parse_join_string <- function(x) {
    parts <- unlist(strsplit(gsub(" ", "", x, fixed = TRUE), "="))
    if (length(parts) != 2)
        stop("unsupported format for the link description. ",
             "Please see ?linkSampleData")
    a <- unlist(strsplit(parts[1L], ".", fixed = TRUE))
    b <- unlist(strsplit(parts[2L], ".", fixed = TRUE))
    if (length(a) < 2 | length(b) < 2)
        stop("unsupported format for the link desciption. ",
             "Please see ?linkSampleData")
    c(a[1L], paste(a[-1L], collapse = "."),
      b[1L], paste0(b[-1L], collapse = "."))
}

#' Create a link matrix for specified vectors.
#'
#' @return `integer` `matrix` with the indices of the matches.
#'
#' @author Johannes Rainer
#'
#' @importFrom S4Vectors findMatches from to
#'
#' @noRd
#'
#' @examples
#'
#' .link_matrix(1:3, 2:20)
#' .link_matrix(c("a", "a", "b", "d", "b", "c"), c("g", "a", "b", "e"))
.link_matrix <- function(from = integer(), to = integer()) {
    res <- findMatches(from, to)
    cbind(from(res), to(res))
}

#' @description
#'
#' Subset a `LinkedMsExperiment` by sample also subsetting and updating all
#' linked data:
#'
#' Subsetting with `[i, j]`:
#' - support re-ordering (`j = c(4, 2, 3)`).
#' - support duplication (`j = c(1, 1, 2)`). -> make that a special case?
#' - keep only elements matching a sample after subsetting - and any unlinked
#'   element.
#'
#' @note
#'
#' how could we improve the performance of this subsetting? Copying over the
#' whole object sounds like not an ideal thing to do.
#'
#' Maybe have helper functions `splitBySample` to avoid repeatedly copying the
#' original data. Parameter `newx` might help here - but not sure if that's
#' indeed the case.
#'
#' @param x `LinkedMsExperiment`.
#'
#' @param i `integer`
#'
#' @param newx `LinkedMsExperiment`. Result objects. Might help avoiding
#'     repeatedly copying the object if `.extractSamples` is called within
#'     a loop. Also, providing `newx = new("LinkedMsExperiment")` would perform
#'     a *lightweight* extraction, dropping anything which is not linked.
#'
#' @author Johannes Rainer
#'
#' @importFrom methods slot<- callNextMethod
#'
#' @noRd
.extractSamples <- function(x, i, newx = x) {
    if (!nrow(sampleData(x)))
        return(x)
    slot(newx, "sampleData", check = FALSE) <- x@sampleData[i, , drop = FALSE]
    for (link in names(slot(x, "sampleDataLinks"))) {
        lmat <- slot(x, "sampleDataLinks")[[link]]
        subsetBy <- mcols(slot(x, "sampleDataLinks"))[link, "subsetBy"]
        idxs <- split(lmat[, 2], as.factor(lmat[, 1]))
        idxs <- idxs[as.character(i)]
        ls <- lengths(idxs)
        idxs <- unlist(idxs, use.names = FALSE)
        element <- .get_element(x, link)
        newx <- .set_element(newx, link, .subset_dim(element, idxs, subsetBy))
        newx@sampleDataLinks[[link]] <- cbind(rep(seq_along(ls), ls),
                                              seq_len(sum(ls)))
        ## Note: keeping also empty lmat - to keep info that there was a link
    }
    newx
}

#' @export
#'
#' @param object An instance of class `MsExperiment`
#'
#' @rdname MsExperiment
experimentFiles  <- function(object) {
    stopifnot(inherits(object, "MsExperiment"))
    object@experimentFiles
}

#' @export
#'
#' @param value An object of the appropriate class for the slot to be
#'     populated.
#'
#' @rdname MsExperiment
"experimentFiles<-" <- function(object, value) {
    stopifnot(inherits(value, "MsExperimentFiles"))
    stopifnot(inherits(object, "MsExperiment"))
    object@experimentFiles <- value
    object
}

#' @export
#'
#' @param object An instance of class `MsExperiment`
#'
#' @rdname MsExperiment
sampleData  <- function(object) {
    stopifnot(inherits(object, "MsExperiment"))
    object@sampleData
}

#' @export
#'
#' @param value An object of the appropriate class for the slot to be
#'     populated.
#'
#' @rdname MsExperiment
"sampleData<-" <- function(object, value) {
    stopifnot(inherits(value, "DataFrame"))
    stopifnot(inherits(object, "MsExperiment"))
    object@sampleData <- value
    object
}

#' Simple helper function that returns the number of elements (for the selected
#' dimension).
#'
#' @noRd
.nelements <- function(x, dim = 1L) {
    dims <- dim(x)
    if (length(dims))
        dims[dim]
    else length(x)
}

#' Helper to subset `x` to elements `i` on dimension `subsetBy`. If `x` has
#' no dimensions `x[i]` is returned, if it has dimensions and `subsetBy == 1L`
#' `x[i, ]` is returned and if `subsetBy == 2L` `x[, i]` is returned.
#'
#' @author Johannes Rainer
#'
#' @noRd
.subset_dim <- function(x, i, subsetBy = 1L) {
    if (length(dim(x))) {
        if (subsetBy == 1L)
            x <- x[i, , drop = FALSE]
        if (subsetBy == 2L)
            x <- x[, i, drop = FALSE]
    } else x <- x[i]
    x
}

#' @export
#'
#' @rdname MsExperiment
qdata <- function(object) {
    stopifnot(inherits(object, "MsExperiment"))
    object@qdata
}

#' @export
#'
#' @rdname MsExperiment
"qdata<-" <- function(object, value) {
    stopifnot(inherits(object, "MsExperiment"))
    object@qdata <- value
    object
}

.ms_experiment_is_empty <- function(object) {
    if (length(object@experimentFiles)) return(FALSE)
    if (length(object@spectra)) return(FALSE)
    if (length(object@qdata)) return(FALSE)
    if (length(object@otherData)) return(FALSE)
    if (length(object@sampleData)) return(FALSE)
    TRUE
}

#' @title Import MS spectra data of an experiment
#'
#' @description
#'
#' Read/import MS spectra data of an experiment from the respective (raw)
#' data files into an [MsExperiment()] object. Files provided with the
#' `spectraFiles` parameter are imported as a `Spectra` object and each
#' file is automatically *linked* to rows (samples) of a `sampleData`
#' data frame (if provided).
#'
#' @param spectraFiles `character` with the (absolute) file names of the MS
#'     data files that should be imported as a [Spectra::Spectra()] object.
#'
#' @param sampleData `data.frame` or `DataFrame` with the sample annotations.
#'     Each row is expected to contain annotations for one file (sample). The
#'     order of the data frame's rows is expected to match the order of the
#'     provided files (with parameter `spectraFiles`).
#'
#' @param ... additional parameters for the [Spectra::Spectra()] call to
#'     import the data.
#'
#' @return `MsExperiment`.
#'
#' @author Johannes Rainer
#'
#' @importMethodsFrom Spectra Spectra
#'
#' @export
#'
#' @examples
#'
#' ## Define the files of the experiment to import. For our example we get
#' ## the files from the *MsDataHub*
#' library(MsDataHub)
#' fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
#'          MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
#'
#' ## Define a data frame with some sample annotations
#' ann <- data.frame(
#'     injection_index = c(1, 9),
#'     sample_id = c("QC_1", "QC_3"))
#'
#' ## Import the data
#' library(MsExperiment)
#' mse <- readMsExperiment(spectraFiles = fls, ann)
#' mse
#'
#' ## Access the spectra data
#' spectra(mse)
#'
#' ## Access the sample annotations
#' sampleData(mse)
#'
#' ## Import the data reading all MS spectra directly into memory
#' mse <- readMsExperiment(spectraFiles = fls, ann,
#'     backend = Spectra::MsBackendMemory())
#' mse
readMsExperiment <- function(spectraFiles = character(),
                             sampleData = data.frame(), ...) {
    spectraFiles <- normalizePath(spectraFiles)
    if (!nrow(sampleData))
        sampleData <- data.frame(sample_index = seq_along(spectraFiles))
    if (nrow(sampleData) != length(spectraFiles))
        stop("Number of rows in 'sampleData' have to match the number of ",
             "files in parameter 'spectraFiles'")
    sampleData$spectraOrigin <- spectraFiles
    if (!inherits(sampleData, "DataFrame"))
        sampleData <- DataFrame(sampleData)
    if (!(is.null(rownames(sampleData)) ||
            all(rownames(sampleData) == basename(spectraFiles)))) {
        stop("'rownames(sampleData)' must be NULL or equal to ",
             "'basename(spectraFiles)'.")
    }
    rownames(sampleData) <- basename(spectraFiles)
    x <- MsExperiment()
    sampleData(x) <- sampleData
    spectra(x) <- Spectra(spectraFiles, ...)
    linkSampleData(x, with = "sampleData.spectraOrigin = spectra.dataOrigin")
}

#' @title Consolidate links between samples and spectra after filtering
#'
#' @description
#'
#' If @spectra got filtered possibly present *links* between them and samples
#' will no longer be valid and need to be updated/fixed. This function
#' consolidates these links using a spectra variable `"._SPECTRA_IDX"` in
#' `@spectra` that needs to represent/contain the index of the spectra
#' **before** filtering.
#'
#' @param x `MsExperiment`
#'
#' @author Johannes Rainer
#' @noRd
.update_sample_data_links_spectra <- function(x) {
    sdl <- .sample_data_links(x, "spectra")[[1L]]
    idx <- match(sdl[, 2L], x@spectra$._SPECTRA_IDX)
    keep <- !is.na(idx)
    sdl <- sdl[keep, , drop = FALSE]
    sdl[, 2L] <- idx[keep]
    x@sampleDataLinks[["spectra"]] <- sdl
    x
}

#' @export
#'
#' @rdname MsExperiment
spectraSampleIndex <- function(x, duplicates = c("first", "keep")) {
    duplicates <- match.arg(duplicates)
    if (duplicates == "first") {
        .spectra_sample_index_first(x@sampleDataLinks[["spectra"]],
                                    length(x@spectra))
    } else {
        .spectra_sample_index_all(x@sampleDataLinks[["spectra"]],
                                  length(x@spectra))
    }
}

#' Return an `integer` vector with the sample index for each spectrum. If
#' a spectrum is assigned to more than one sample, the index of the first
#' is returned and a warning shown. For spectra without a sample assignment
#' `NA_integer_` is returned.
#'
#' @param x 2 column `matrix`, first column being sample index, second
#'     spectra index.
#'
#' @param nspectra length of the object's `Spectra` object
#'
#' @return `integer` of length equal to the number of spectra.
#'
#' @noRd
.spectra_sample_index_first <- function(x, nspectra) {
    if (length(x)) {
        if (anyDuplicated(x[, 2L])) {
            warning("Found at least one spectrum that is assigned to more ",
                    "than one sample. Will return the first sample for these. ",
                    "Consider using 'duplicates = \"all\"' to retrieve all ",
                    "mappings.")
            res <- x[match(seq_len(nspectra), x[, 2L]), 1L]
        } else {
            res <- rep(NA_integer_, nspectra)
            res[x[, 2L]] <- x[, 1L]
        }
    } else res <- rep(NA_integer_, nspectra)
    res
}

#' Return an `list` of integer vectors with the sample indices for each
#' spectrum. For spectra without a sample assignment `integer()` is returned.
#'
#' @param x 2 column `matrix`, first column being sample index, second
#'     spectra index.
#'
#' @param nspectra length of the object's `Spectra` object
#'
#' @return `list` of length equal to the number of spectra.
#'
#' @noRd
.spectra_sample_index_all <- function(x, nspectra) {
    if (length(x)) {
        res <- split(x[, 1L], f = factor(x[, 2L], levels = seq_len(nspectra)))
    } else res <- replicate(nspectra, integer())
    unname(res)
}

test_that(".sample_data_links works", {
    x <- new("MsExperiment")
    expect_equal(
        .sample_data_links(x),
        new("SimpleList", elementMetadata = DataFrame(subsetBy = integer())))

    x@sampleDataLinks$a <- cbind(1:2, 2:3)
    x@sampleDataLinks$b <- cbind(1:4, 1:4)

    res <- .sample_data_links(x, "a")
    expect_equal(res[[1L]], cbind(1:2, 2:3))

    res <- .sample_data_links(x, "d")
    expect_true(length(res) == 0)
})

test_that(".valid_link works", {
    expect_error(.valid_link(3), "needs to be")
    expect_error(.valid_link(matrix(ncol = 2, nrow = 4, "a")), "needs to be")
    expect_error(.valid_link(matrix(ncol = 3, nrow = 2, 1L)), "2 columns")
    mat <- cbind(1:3, 1L)
    expect_error(.valid_link(mat, nfrom = 1), " <= 1")
    mat[, 2] <- -1L
    expect_error(.valid_link(mat, nfrom = 4), "smaller than 1")
})

test_that(".add_sample_data_link works", {
    x <- new("MsExperiment")
    mat <- cbind(1:3, 1L)
    expect_equal(.add_sample_data_link(x, mat), x)

    sampleData(x) <- DataFrame(df)
    x@metadata$test <- 1:5
    expect_error(.add_sample_data_link(x, mat, with = "metadata.test"), "<= 2")
    expect_error(.add_sample_data_link(x, mat[1:2, ], "spectra"), "empty")

    x@metadata$test_char <- c("a", "b", "c")
    mat <- mat[1:2, ]
    mat[, 2] <- c(3L, 1L)
    res <- .add_sample_data_link(x, mat, with = "metadata.test_char")
    expect_true(names(res@sampleDataLinks) == "metadata.test_char")
    expect_equal(mcols(res@sampleDataLinks)["metadata.test_char", "subsetBy"],
                 1L)

    res@metadata$other_var = c("b")
    mat <- cbind(1L, 1L)
    res <- .add_sample_data_link(res, mat, with = "metadata.other_var",
                                 subsetBy = 2L)
    expect_equal(mcols(res@sampleDataLinks)["metadata.other_var", "subsetBy"],
                 2L)
})

test_that(".get_element works", {
    res <- .get_element(mse, "sampleData.mzML_file")
    expect_equal(res, sampleData(mse)$mzML_file)
    res <- .get_element(mse, "sampleData.dont_exist")
    expect_equal(res, NULL)

    expect_error(.get_element(mse, "noslot"), "No slot named")

    mse@metadata$new_entry.dot <- "value"
    res <- .get_element(mse, "metadata.new_entry.dot")
    expect_equal(res, "value")

    res <- .get_element(mse, "sampleData")
    expect_equal(res, sampleData(mse))

    res <- .get_element(mse, "experimentFiles.other_file")
    expect_equal(res, "other_file.txt")
})

test_that(".set_element works", {
    tmp <- new("MsExperiment")
    expect_error(.set_element(tmp, "does_not_exist", 4), "No slot")
    res <- .set_element(tmp, "metadata.new_val", data.frame(id = 1:3))
    expect_true(names(metadata(res)) == "new_val")
    expect_equal(metadata(res)$new_val, data.frame(id = 1:3))

    res <- .set_element(tmp, "spectra", spectra(mse)[1:3])
    expect_equal(length(spectra(res)), 3)
    res <- .set_element(res, "spectra.new_val", "a")
    expect_true(all(spectra(res)$new_val == "a"))
})

test_that(".parse_join_string works", {
    res <- .parse_join_string("spectra.dataOrigin =  otherData.something")
    expect_equal(res, c("spectra", "dataOrigin", "otherData", "something"))
    expect_error(.parse_join_string("does not work"), "unsupported format")
})

test_that(".link_matrix works", {
    res <- .link_matrix()
    expect_true(is.matrix(res))
    expect_true(nrow(res) == 0)

    res <- .link_matrix(c("a", "a", "b", "d", "b", "c"), c("g", "a", "b", "e"))
    expect_equal(res, cbind(c(1, 2, 3, 5), c(2, 2, 3, 3)))

    a <- c("d", "a", "a", "b", "e", "b", "d", "d")
    b <- c("b", "a", "a", "c", "d", "a")
    exp <- cbind(c(1, 2, 2, 2, 3, 3, 3, 4, 6, 7, 8),
                 c(5, 2, 3, 6, 2, 3, 6, 1, 1, 5, 5))
    res <- .link_matrix(a, b)
    expect_equal(res, exp)

    exp <- cbind(c(1, 1, 2, 2, 3, 3, 5, 5, 5, 6, 6),
                 c(4, 6, 2, 3, 2, 3, 1, 7, 8, 2, 3))
    res <- .link_matrix(b, a)
    expect_equal(res, exp)
})

test_that(".extractSamples works", {
    tmp <- mse
    res <- .extractSamples(tmp, i = 2)
    expect_equal(spectra(res), spectra(tmp))
    expect_equal(sampleData(res), sampleData(tmp)[2, ])

    ## Establish links.
    ## n:1 mapping
    tmp <- linkSampleData(tmp, "experimentFiles.other_file",
                          sampleIndex = c(1, 2), withIndex = c(1, 1))
    ## 1:1 mapping
    tmp <- linkSampleData(tmp, "experimentFiles.mzML_file",
                          sampleIndex = c(1, 2), withIndex = c(1, 2))
    ## 1:n mapping
    tmp <- linkSampleData(tmp, "spectra",
                          sampleIndex = match(basename(spectra(mse)$dataOrigin),
                                              sampleData(mse)$mzML_file),
                          withIndex = seq_along(spectra(mse)))
    ## n:m mapping
    metadata(tmp)[["multivals"]] <- c("AB", "A", "B")
    tmp <- linkSampleData(tmp, "metadata.multivals",
                          sampleIndex = c(1, 1, 2, 2),
                          withIndex = c(1, 2, 1, 3))
    ## link a data.frame
    metadata(tmp)$df <- data.frame(a = 1:3, b = 5:7)
    tmp <- linkSampleData(tmp, "metadata.df", sampleIndex = c(1, 2),
                          withIndex = c(1, 2))
    ## Add a non-linked element
    metadata(tmp)$not_linked <- "not linked"

    ## Extract a single sample
    res <- .extractSamples(tmp, i = 2)
    expect_equal(sampleData(res), sampleData(tmp)[2, ])
    expect_equal(experimentFiles(res)$other_file, "other_file.txt")
    expect_equal(experimentFiles(res)$mzML_file, experimentFiles(tmp)$mzML_file[2])
    expect_equal(spectra(res), spectra(tmp)[932:1862])
    expect_equal(metadata(res)$multivals, c("AB", "B"))
    expect_equal(metadata(res)$df, metadata(tmp)$df[2, , drop = FALSE])
    expect_equal(metadata(res)$not_linked, "not linked")

    ## Re-ordering of samples
    res <- .extractSamples(tmp, i = c(2, 1))
    expect_equal(sampleData(res), sampleData(tmp)[2:1, ])
    ## ! n:1 mapping gets duplicated!
    expect_equal(experimentFiles(res)$other_file, rep("other_file.txt", 2))
    ## 1:n mapping are fine
    expect_equal(experimentFiles(res)$mzML_file,
                 experimentFiles(tmp)$mzML_file[2:1])
    expect_equal(rtime(spectra(res))[932:1862], rtime(spectra(tmp))[1:931])
    expect_equal(rtime(spectra(res))[1:931], rtime(spectra(tmp))[932:1862])
    expect_equal(metadata(res)$df, metadata(tmp)$df[2:1, , drop = FALSE])
    ## ! n:m mapping gets duplicated!
    expect_equal(metadata(res)$multivals, c("AB", "B", "AB", "A"))

    ## Duplication of samples
    res <- .extractSamples(tmp, i = c(2, 2))
    expect_equal(sampleData(res), sampleData(tmp)[c(2, 2), ])
    expect_equal(experimentFiles(res)$other_file, rep("other_file.txt", 2))
    expect_equal(experimentFiles(res)$mzML_file,
                 experimentFiles(tmp)$mzML_file[c(2, 2)])
    expect_equal(spectra(res)[1:931], spectra(tmp)[932:1862])
    expect_equal(spectra(res)[932:1862], spectra(tmp)[932:1862])
    expect_equal(metadata(res)$multivals, c("AB", "B", "AB", "B"))
    expect_equal(metadata(res)$df, metadata(tmp)$df[c(2, 2), , drop = FALSE])
    expect_equal(metadata(res)$not_linked, "not linked")

    ## Link a SummarizedExperiment (by column).
    library(SummarizedExperiment)
    se <- SummarizedExperiment(
        cbind(QC2 = 1:10, QC3 = 11:20, QC1 = 21:30),
        colData = DataFrame(sample = c("QC2", "QC3", "QC1"), idx = 1:3))
    tmp@qdata <- se
    tmp <- linkSampleData(tmp, with = "sampleData.sample = qdata.sample")
    expect_equal(mcols(tmp@sampleDataLinks["qdata"])$subsetBy, 2L)

    res <- .extractSamples(tmp, i = 2)
    expect_equal(sampleData(res), sampleData(tmp)[2, ])
    expect_equal(res@qdata$sample, "QC2")
    expect_equal(assay(res@qdata),
                 matrix(1:10, ncol = 1, dimnames = list(character(), "QC2")))

    res <- .extractSamples(tmp, i = c(1, 1))
    expect_equal(res@qdata$sample, c("QC1", "QC1"))
})

test_that("experimentFiles works", {
    expect_error(experimentFiles(4))
    res <- experimentFiles(MsExperiment())
    expect_true(length(res) == 0)

    res <- experimentFiles(mse)
    expect_s4_class(res, "MsExperimentFiles")
    expect_true(length(res) == 2)
})

test_that("experimentFiles<- works", {
    expect_error(experimentFiles(4) <- 4)
    m <- MsExperiment()
    expect_error(experimentFiles(m) <- c(a = "b"))
    mf <- MsExperimentFiles(a = "b")
    experimentFiles(m) <- mf
    expect_equal(experimentFiles(m), mf)
})

test_that("sampleData works", {
    expect_error(sampleData(4))
    res <- sampleData(MsExperiment())
    expect_s4_class(res, "DataFrame")
    expect_true(nrow(res) == 0)

    res <- sampleData(mse)
    expect_s4_class(res, "DataFrame")
    expect_true(nrow(res) > 0)
})

test_that("sampleData<- works", {
    expect_error(sampleData(4) <- DataFrame())
    m <- MsExperiment()
    expect_error(sampleData(m) <- 4)
    df <- DataFrame(a = 1:3, b = 3:1)
    sampleData(m) <- df
    expect_equal(sampleData(m), df)
})

test_that(".nelements works", {
    expect_equal(.nelements(1:10), 10)
    expect_equal(.nelements(matrix(nrow = 3, ncol = 2)), 3)
    expect_equal(.nelements(matrix(nrow = 3, ncol = 2), 2), 2)
})

test_that(".subset_dim works", {
    res <- .subset_dim(1:10, c(2, 3))
    expect_equal(res, 2:3)
    res <- .subset_dim(cbind(1:10, 11:20), c(2, 3))
    expect_equal(res, cbind(2:3, 12:13))
    res <- .subset_dim(cbind(1:10, 11:20), 2, subsetBy = 2)
    expect_equal(res, matrix(11:20, ncol = 1))
})

test_that("qdata, qdata<- works", {
    expect_error(qdata(3), "MsExperiment")

    expect_equal(qdata(mse), NULL)

    tmp <- MsExperiment()
    se <- SummarizedExperiment::SummarizedExperiment(matrix(nrow = 3, ncol = 4))
    expect_error(qdata(se) <- 4, "MsExperiment")
    qdata(tmp) <- se
    expect_equal(qdata(tmp), se)
})

test_that(".ms_experiment_is_empty works", {
    expect_true(.ms_experiment_is_empty(MsExperiment()))
    expect_false(.ms_experiment_is_empty(mse))
})

test_that("readMsExperiment works", {
    expect_warning(a <- readMsExperiment(), "no")
    expect_s4_class(a, "MsExperiment")
    expect_true(length(a) == 0)
    expect_true(length(spectra(a)) == 0)
    expect_error(a <- readMsExperiment("a"), "not found")
    fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
             MsDataHub::X20171016_POOL_POS_3_105.134.mzML())

    a <- readMsExperiment(fls)
    expect_s4_class(a, "MsExperiment")
    expect_true(length(a) == 2)
    expect_true(nrow(sampleData(a)) == 2)

    df <- data.frame(sidx = 1:3, other_ann = c("a", "b", "c"))
    expect_error(readMsExperiment(fls, df), "of files")
    a <- readMsExperiment(fls[1:2], df[1:2, ])
    expect_s4_class(a, "MsExperiment")
    expect_true(length(a) == 2)
    expect_true(nrow(sampleData(a)) == 2)
    expect_equal(sampleData(a)$other_ann, c("a", "b"))
})

test_that(".update_sample_data_links_spectra works", {
    a <- readMsExperiment(fls)
    tmp <- a
    tmp@spectra$._SPECTRA_IDX <- seq_along(tmp@spectra)
    tmp@spectra <- tmp@spectra[c(5, 14, 1000, 2, 200)]
    res <- .update_sample_data_links_spectra(tmp)
    expect_equal(res@sampleDataLinks[["spectra"]][, 1L], c(1L, 1L, 1L, 1L, 2L))
    expect_equal(res@sampleDataLinks[["spectra"]][, 2L], c(4L, 1L, 2L, 5L, 3L))
    expect_equal(res@spectra$scanIndex,
                 a@spectra$scanIndex[c(5, 14, 1000, 2, 200)])
    expect_equal(res@sampleData, tmp@sampleData)

    expect_true(length(spectra(res[1L])) == 4)
    expect_true(length(spectra(res[2L])) == 1)
})

test_that(".spectra_sample_index_first works", {
    res <- .spectra_sample_index_first(matrix(nrow = 0, ncol = 2), 100)
    expect_equal(length(res), 100)
    expect_equal(res, rep(NA_integer_, 100))

    a <- cbind(c(1, 1, 1, 1, 2, 2, 2, 2), 4:11)
    res <- .spectra_sample_index_first(a, 20)
    expect_equal(length(res), 20)
    expect_equal(res[4:11], c(1, 1, 1, 1, 2, 2, 2, 2))
    expect_true(all(is.na(res[12:20])))

    a <- cbind(c(1, 1, 1, 1, 2, 2, 2, 2), c(4:10, 5))
    expect_warning(res <- .spectra_sample_index_first(a, 20), "Found at least")
    expect_equal(length(res), 20)
    expect_equal(res[4:10], c(1, 1, 1, 1, 2, 2, 2))
    expect_true(all(is.na(res[11:20])))
})

test_that(".spectra_sample_index_all works", {
    res <- .spectra_sample_index_all(matrix(nrow = 0, ncol = 2), 100)
    expect_equal(length(res), 100)
    expect_true(is.list(res))
    expect_true(all(lengths(res) == 0))
    expect_equal(res, replicate(100, integer()))

    a <- cbind(c(1, 1, 1, 1, 2, 2, 2, 2), 4:11)
    res <- .spectra_sample_index_all(a, 20)
    expect_equal(length(res), 20)
    expect_equal(res[4:11], list(1L, 1L, 1L, 1L, 2L, 2L, 2L, 2L))
    expect_true(all(lengths(res[12:20]) == 0))

    a <- cbind(c(1, 1, 1, 1, 2, 2, 2, 2), c(4:10, 5))
    res <- .spectra_sample_index_all(a, 20)
    expect_equal(length(res), 20)
    expect_equal(res[4:10], list(1L, c(1L, 2L), 1L, 1L, 2L, 2L, 2L))
    expect_true(all(lengths(res[11:20]) == 0))
})

test_that("spectraSampleIndex works", {
    a <- MsExperiment()
    expect_equal(spectraSampleIndex(a), integer())

    a <- mse
    expect_equal(spectraSampleIndex(a), rep(NA_integer_, length(spectra(a))))

    a@sampleDataLinks[["spectra"]] <- cbind(c(1L, 2L, 1L, 1L),
                                            c(132L, 2L, 342L, 54L))
    res <- spectraSampleIndex(a)
    expect_equal(length(res), length(spectra(a)))
    expect_true(sum(!is.na(res)) == 4L)
    expect_equal(res[!is.na(res)], c(2L, 1L, 1L, 1L))
    expect_equal(res[c(132, 2, 342, 54)], c(1, 2, 1, 1))

    res <- spectraSampleIndex(a, duplicates = "keep")
    expect_equal(res[c(132, 2, 342, 54)], list(1L, 2L, 1L, 1L))

    a@sampleDataLinks[["spectra"]] <- cbind(c(1L, 2L, 1L, 1L, 2L),
                                            c(132L, 2L, 342L, 54L, 2L))
    expect_warning(res <- spectraSampleIndex(a), "Found at least")
    expect_equal(res[c(132, 2, 342, 54)], c(1L, 2L, 1L, 1L))

    res <- spectraSampleIndex(a, duplicates = "keep")
    expect_equal(res[c(132, 2, 342, 54)], list(1L, c(2L, 2L), 1L, 1L))
})

test_that("linkSampleData,MsExperiment works", {
    res <- linkSampleData(mse)
    expect_true(length(res@sampleDataLinks) == 0)

    expect_error(linkSampleData(mse, with = "experimentFile.fls"),
                 "unsupported")
    expect_error(
        linkSampleData(mse, with = "experimentFiles.mzML_file", withIndex = 1),
        "Length")
    expect_error(linkSampleData(
        mse, with = "experimentFile.mzML_file", withIndex = c(2, 1)), "No slot")
    expect_error(linkSampleData(
        mse, with = "experimentFiles.mzML_files", withIndex = c(2, 1)), "empty")

    res <- linkSampleData(mse, with = "experimentFiles.mzML_file",
                          withIndex = c(2, 1))
    expect_true(length(res@sampleDataLinks) == 1)
    expect_equal(res@sampleDataLinks[["experimentFiles.mzML_file"]],
                 cbind(1:2, 2:1))
    expect_warning(linkSampleData(
        res, with = "experimentFiles.mzML_file", withIndex = 1:2),
        "Overwriting")

    ## link all sample to one file.
    res <- linkSampleData(mse, with = "experimentFiles.mzML_file",
                          withIndex = c(1, 1))
    expect_equal(res@sampleDataLinks[["experimentFiles.mzML_file"]],
                 cbind(1:2, c(1L, 1L)))

    ## link one sample to one file.
    res <- linkSampleData(mse, with = "experimentFiles.mzML_file",
                          sampleIndex = 1L, withIndex = 2L)
    expect_equal(res@sampleDataLinks[["experimentFiles.mzML_file"]],
                 cbind(1L, 2L))

    ## link using a SQL-like expression
    expect_error(linkSampleData(mse, with = "bla.blu = ble.blo"),
                 "one of the slot")

    sampleData(mse)$orgfile <- unique(spectra(mse)$dataOrigin)
    res <- linkSampleData(mse, with = "sampleData.orgfile = spectra.dataOrigin")
    res2 <- linkSampleData(mse, with = "spectra.dataOrigin = sampleData.orgfile")
    expect_equal(res@sampleDataLinks[[1L]], res2@sampleDataLinks[[1L]])

    res3 <- linkSampleData(
        mse, with = "spectra",
        sampleIndex = match(basename(spectra(mse)$dataOrigin),
                            sampleData(mse)$mzML_file),
        withIndex = seq_along(spectra(mse)))
    expect_equal(unname(res@sampleDataLinks[[1L]]),
                 unname(res3@sampleDataLinks[[1L]]))
})

test_that("[,LinkedMsExperiment works", {
    tmp <- mse
    expect_error(res <- tmp[4], "out-of-bound")
    expect_error(res <- tmp[1, 2], "is supported")
    expect_error(res <- tmp[c(TRUE, FALSE, TRUE)], "number of")
    expect_warning(tmp[c("b")], "rownames")

    res <- tmp[c(TRUE, FALSE)]
    expect_equal(sampleData(res), sampleData(tmp)[1, ])

    ## Subset including qdata
    library(SummarizedExperiment)
    sd <- DataFrame(sample = c("QC1", "QC2", "QC3"), idx = c(1, 3, 5))
    se <- SummarizedExperiment(colData = sd, assay = cbind(1:10, 11:20, 21:30))

    mse2 <- mse
    qdata(mse2) <- se

    res <- mse2[2]
    expect_equal(length(res), 1L)
    expect_equal(sampleData(res), sampleData(mse2)[2, ])
    expect_equal(spectra(res), spectra(mse2))
    expect_equal(qdata(res), qdata(mse2))

    ## Link spectra
    spectra(mse2)$mzML_file <- basename(spectra(mse2)$dataOrigin)
    mse2 <- linkSampleData(
        mse2, with = "sampleData.mzML_file = spectra.mzML_file")
    res <- mse2[2]
    expect_equal(
        spectra(res),
        spectra(mse2)[spectra(mse2)$mzML_file == basename(fls)[2L]])
    res <- mse2[1]
    expect_equal(
        spectra(res),
        spectra(mse2)[spectra(mse2)$mzML_file == basename(fls)[1L]])

    ## Link experiment files
    mse2 <- linkSampleData(mse2, with = "experimentFiles.mzML_file",
                           sampleIndex = c(1, 2), withIndex = c(1, 2))
    res <- mse2[2]
    expect_equal(experimentFiles(res)[["mzML_file"]],
                 experimentFiles(mse2)[["mzML_file"]][2L])
    res <- mse2[1]
    expect_equal(experimentFiles(res)[["mzML_file"]],
                 experimentFiles(mse2)[["mzML_file"]][1L])

    ## Link qdata
    mse2 <- linkSampleData(mse2, with = "sampleData.sample = qdata.sample")
    res <- mse2[2]
    expect_equal(qdata(res), qdata(mse2)[, 2L])
    res <- mse2[1]
    expect_equal(qdata(res), qdata(mse2)[, 1L])

    ##
    qdata(mse2) <- se[, 1:2]
    res <- mse2[1:2]
    expect_equal(res, mse2)

    ## Subsetting with negative indices
    msel <- mse
    spectra(msel)$mzML_file <- basename(spectra(msel)$dataOrigin)
    msel <- linkSampleData(
        msel, with = "sampleData.mzML_file = spectra.mzML_file")
    res <- msel[-1]
    expect_equal(length(res), (length(msel) - 1))
    expect_equal(sampleData(res)$sample, sampleData(msel)$sample[-1])
    expect_equal(unique(basename(spectra(res)$dataOrigin)),
                 unique(basename(spectra(msel)$dataOrigin))[2])

    expect_error(msel[c(1, -2)], "not supported")

    ## Multiple files/samples.
    sd <- data.frame(sample = c("A", "B", "C"))
    f <- c(fls, MsDataHub::PestMix1_DDA.mzML())
    msel <- readMsExperiment(spectraFiles = f, sampleData = sd)
    res <- msel[-2]
    expect_true(length(res) == 2L)
    expect_equal(sampleData(res)$sample, c("A", "C"))
    expect_equal(unique(basename(spectra(res)$dataOrigin)),
                 unique(basename(spectra(msel)$dataOrigin))[c(1, 3)])

    ref <- msel[2]
    res <- msel[c(-1, -3)]
    expect_equal(sampleData(ref), sampleData(res))
    expect_equal(spectra(ref), spectra(res))
    expect_equal(ref@sampleDataLinks, res@sampleDataLinks)
})

test_that("MsExperiment works", {
    m <- MsExperiment()
    expect_s4_class(m, "MsExperiment")

    sdata <- data.frame(sample_name = c("A", "B"), sample_index = c(4, 12),
                        sample_file = basename(fls))
    s <- Spectra(fls)
    s$file <- basename(dataOrigin(s))

    m <- MsExperiment(spectra = s)
    expect_equal(spectra(m), s)

    m <- MsExperiment(spectra = s, sampleData = sdata)
    expect_equal(spectra(m), s)
    expect_equal(sampleData(m), DataFrame(sdata))

    s_sql <- tempfile()
    s <- setBackend(s, MsBackendOfflineSql(), drv = RSQLite::SQLite(),
                    dbname = s_sql, BPPARAM = SerialParam())

    m <- MsExperiment(spectra = s)
    expect_equal(spectra(m), s)
    expect_equal(sampleData(m), DataFrame())

    m <- MsExperiment(spectra = s, sampleData = sdata)
    dbWriteSampleData(m)

    ## Read sample data from database
    m <- MsExperiment(spectra = s)
    expect_equal(spectra(m), s)
    expect_equal(sampleData(m)[, 1:3], DataFrame(sdata))

    m <- MsExperiment(spectra = s, sampleData = sdata)
    expect_equal(sampleData(m), DataFrame(sdata))

    m <- linkSampleData(m, with = "sampleData.sample_file = spectra.file")
    dbWriteSampleData(m)
    m2 <- MsExperiment(spectra = s)
    expect_equal(sampleData(m2)[, 1:3], DataFrame(sdata))
    expect_equal(m@sampleDataLinks[["spectra"]],
                 m2@sampleDataLinks[["spectra"]])
})

test_that("show,MsExperiment works", {
    expect_output(show(MsExperiment()), "MsExperiment")
    expect_output(show(mse), "Experiment data")
    expect_output(show(MsExperiment()), "Empty object")
    otherData(mse)$X <- 1
    expect_output(show(mse), "Other data: X")
})

test_that("metadata<-,metadata,MsExperiment works", {
    m <- MsExperiment()
    metadata(m) <- list(version = "1.2", data = "1900")
    res <- metadata(m)
    expect_equal(res$version, "1.2")
    expect_equal(res$data, "1900")
})

test_that("spectra<-,spectra,MsExperiment works", {
    m <- MsExperiment()
    expect_null(spectra(m))

    res <- spectra(mse)
    expect_s4_class(res, "Spectra")
    expect_true(length(res) > 0)

    spectra(m) <- res
    expect_equal(spectra(m), res)

    expect_error(spectra(m) <- "b")
})


test_that("otherData<-,otherData,MsExperiment works", {
    m <- MsExperiment()
    expect_identical(length(otherData(m)), 0L)
    otherData(m)$NUM <- 1
    expect_identical(length(otherData(m)), 1L)
    expect_identical(names(otherData(m)), "NUM")
    expect_identical(otherData(m)[["NUM"]], 1)
    otherData(m)[["NUM"]] <- NULL
    expect_identical(length(otherData(m)), 0L)
})

test_that("filterSpectra,MsExperiment,function works", {
    ## empty object
    res <- filterSpectra(MsExperiment(), filterMsLevel, 2L)
    expect_s4_class(res, "MsExperiment")
    expect_true(length(res) == 0L)

    ## an object without links.
    res <- filterSpectra(mse, filterMsLevel, 2L)
    expect_s4_class(res, "MsExperiment")
    expect_equal(sampleData(res), sampleData(mse))
    expect_true(length(spectra(res)) == 0L)
    expect_equal(res@sampleDataLinks, mse@sampleDataLinks)

    ## an object with links between samples and spectra
    tmp <- readMsExperiment(fls)
    res <- filterSpectra(tmp, filterMsLevel, 2L)
    expect_s4_class(res, "MsExperiment")
    expect_equal(sampleData(res), sampleData(tmp))
    expect_true(length(spectra(res)) == 0L)
    expect_equal(names(res@sampleDataLinks), names(tmp@sampleDataLinks))
    expect_true(nrow(res@sampleDataLinks[["spectra"]]) == 0L)

    ## Just reducing.
    res <- filterSpectra(tmp, filterRt, c(200, 210))
    expect_s4_class(res, "MsExperiment")
    expect_equal(sampleData(res), sampleData(tmp))
    expect_true(all(rtime(spectra(res)) >= 200 & rtime(spectra(res)) <= 210))
    ## check that sample/spectra links are valid
    a <- spectra(res[1L])
    ref <- filterRt(spectra(tmp[1L]), c(200, 210))
    expect_equal(a$scanIndex, ref$scanIndex)
    expect_equal(a$dataOrigin, ref$dataOrigin)
    a <- spectra(res[2L])
    ref <- filterRt(spectra(tmp[2L]), c(200, 210))
    expect_equal(a$scanIndex, ref$scanIndex)
    expect_equal(a$dataOrigin, ref$dataOrigin)
})

test_that("readMsExperiment checks that sample names match", {
    df <- data.frame(
        row.names = c("sample1", "sample2")
    )
    # Error: rownames(df) do not match with basename(fls)
    res <- readMsExperiment(spectraFiles = fls, sampleData = df) |>
        expect_error()
    # No error: rownames of sample metadata file match with basename(files)
    rownames(df) <- basename(fls)
    res <- readMsExperiment(spectraFiles = fls, sampleData = df) |>
        expect_no_error()
    # No error: rownames are not included, the function expects that user takes
    # care that the order of sample metadata matches with files.
    rownames(df) <- NULL
    res <- readMsExperiment(spectraFiles = fls, sampleData = df) |>
        expect_no_error()
})

#' Functionality to store/retrieve MsExperiment data to/from databases.
#'
#' @noRd
NULL

#' @title Write sample annotations to a MsBackendSql SQL database
#'
#' @description
#'
#' For `MsExperiment` objects with their MS data represented by a `Spectra`
#' object that use a `MsBackendSql` backend, its sample annotations can be
#' written to the backend's SQL database with the `dbWriteSampleData()`
#' function.
#' The content of the object's `[sampleData()]` (as well as eventually present
#' *linking* between samples and spectra) will be stored in two separate
#' database tables *sample_data* and *sample_to_msms_spectrum* in the same
#' database.
#'
#' This requires that the MS data of the experiment is *represented* by a
#' `MsBackendSql` backend (see help on the `createMsBackendSqlDatabase` or the
#' *MsBackendSql* package vignette for more information on how to create or
#' use such SQL databases).
#'
#' @param x `MsExperiment` from which sample annotations should be written
#'     to the database.
#'
#' @author Johannes Rainer, Laurent Gatto
#'
#' @importFrom DBI dbExistsTable dbRemoveTable dbWriteTable dbDisconnect
#'
#' @export
#'
#' @examples
#'
#' library(MsExperiment)
#'
#' ## Create a MsBackendSql database from two mzML files.
#' ## Connect first to an empty SQLite database (for the example we create
#' ## a database in a temporary file).
#' library(RSQLite)
#' sqlite_db <- tempfile()
#' con <- dbConnect(SQLite(), sqlite_db)
#'
#' ## Define the files from which we import the data. Test files provided from
#' ## the *MsDataHub*
#' fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
#'     MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
#'
#' ## Create a MsBackendSql database containing the full MS data
#' library(MsBackendSql)
#' createMsBackendSqlDatabase(con, fls)
#'
#' ## Note: alternatively it would be possible to first import the MS data
#' ## to a `Spectra` object and then change the backend to a `MsBackendSql`
#' ## using the `setBackend` function.
#'
#' ## Load this data as a `Spectra` object (using a `MsBackendOfflineSql`
#' ## backend)
#' library(Spectra)
#' sps <- Spectra(sqlite_db, source = MsBackendOfflineSql(),
#'     drv = SQLite())
#' sps
#'
#' ## Define sample annotations for the two data files. Adding one column
#' ## `"file"` that contains the file name of the data files.
#' df <- data.frame(sample = c("QC1", "QC2"), file = basename(fls))
#'
#' ## Add a spectra variable `"file"` to the `Spectra` object with
#' ## the raw data files' file names to simplify the linking between
#' ## samples and spectra performed later.
#' sps$file <- basename(dataOrigin(sps))
#'
#' ## Create a MsExperiment with the spectra and sample data.
#' mse <- MsExperiment(spectra = sps, sampleData = df)
#'
#' ## Establish the link (mapping) between samples and spectra
#' ## using the column `"file"` in the `sampleData` and the spectra
#' ## variable `"file"`.
#' mse <- linkSampleData(mse, with = "sampleData.file = spectra.file")
#' mse
#'
#' ## Write sample data (and the sample to spectra mapping) to the
#' ## *MsBackendSql* database.
#' dbWriteSampleData(mse)
#'
#' ## List the tables in the database
#' dbListTables(con)
#'
#' ## Sample data was thus stored to the database.
#' dbGetQuery(con, "select * from sample_data;")
dbWriteSampleData <- function(x) {
    if (!inherits(x, "MsExperiment"))
        stop("'x' is supposed to be an 'MsExperiment'")
    if (!length(spectra(x))) return()
    if (inherits(spectra(x)@backend, "MsBackendSql")) {
        be <- spectra(x)@backend
        if (inherits(be, "MsBackendOfflineSql")) {
            flags <- be@flags
            be@flags <- 2L # fix to ensure getting a writebale connection
            con <- dbconn(be)
            be@flags <- flags
            on.exit(dbDisconnect(con))
        } else con <- dbconn(be)
        if (dbExistsTable(con, "sample_data")) {
            message("Sample data already available in the database. ",
                    "Will overwrite.")
            dbRemoveTable(con, "sample_data")
        }
        if (dbExistsTable(con, "sample_to_msms_spectrum"))
            dbRemoveTable(con, "sample_to_msms_spectrum")
        sdata <- as.data.frame(sampleData(x))
        if (any(colnames(sdata) == "sample_id_"))
            warning("Replacing already present column \"sample_id_\" ",
                    "in 'sampleData'.")
        sdata$sample_id_ <- seq_len(nrow(sdata))
        dbWriteTable(con, name = "sample_data", sdata)
        sdata_links <- x@sampleDataLinks[["spectra"]]
        if (length(sdata_links)) {
            sdata_links <- as.data.frame(sdata_links)
            colnames(sdata_links) <- c("sample_index", "spectrum_index")
            dbWriteTable(con, name = "sample_to_msms_spectrum", sdata_links)
        }
    } else stop("The object's 'Spectra' object does not use a ",
                "'MsBackendSql' backend.")
}

#' @title Get sample data from a MsBackendSql database
#'
#' @description
#'
#' Retrieve sample data and sample to spectra links from a MsBackendSql
#' database, if present. This requires that the `Spectra` `x` uses a
#' `MsBackendSql` or `MsBackendOfflineSql` backend and that the database
#' contains database tables *sample_data* and eventually
#' *sample_to_msms_spectrum*.
#'
#' @param x `Spectra`
#'
#' @author Johannes Rainer
#'
#' @return information from the database or an empty array.
#'
#' @importFrom DBI dbGetQuery dbListTables
#'
#' @importMethodsFrom BiocGenerics dbconn
#'
#' @noRd
.db_get_sample_data <- function(x) {
    if (inherits(x@backend, "MsBackendSql")) {
        con <- dbconn(x@backend)
        if (inherits(x@backend, "MsBackendOfflineSql"))
            on.exit(dbDisconnect(con))
        if (any(dbListTables(con) == "sample_data"))
            dbGetQuery(con, "select * from sample_data")
        else data.frame()
    } else data.frame()
}
.db_get_sample_spectra_link <- function(x) {
    if (inherits(x@backend, "MsBackendSql")) {
        con <- dbconn(x@backend)
        if (inherits(x@backend, "MsBackendOfflineSql"))
            on.exit(dbDisconnect(con))
        if (any(dbListTables(con) == "sample_to_msms_spectrum"))
            unname(as.matrix(
                dbGetQuery(con, "select * from sample_to_msms_spectrum")))
        else matrix(ncol = 2, nrow = 0)
    } else matrix(ncol = 2, nrow = 0)
}

library(MsBackendSql)
library(RSQLite)
ref_sql <- tempfile()
ref_con <- dbConnect(SQLite(), ref_sql)
mse_db <- mse
spectra(mse_db) <- setBackend(spectra(mse_db), MsBackendSql(), dbcon = ref_con)
dbWriteSampleData(mse_db)

test_that("dbWriteSampleData works", {
    expect_error(dbWriteSampleData("a"), "MsExperiment")
    m <- MsExperiment()
    expect_null(dbWriteSampleData(m))
    ## MsExperiment with Spectra but not using MsBackendSql
    expect_error(dbWriteSampleData(mse), "MsBackendSql")

    s <- spectra(mse)
    sql_file <- tempfile()
    con <- dbConnect(SQLite(), sql_file)
    s <- setBackend(s, MsBackendSql(), dbcon = con)
    s$f <- basename(dataOrigin(s))
    spectra(m) <- s
    ## Empty sampleData
    dbWriteSampleData(m)
    expect_equal(dbListTables(con),
                 c("msms_spectrum", "msms_spectrum_peak_blob2", "sample_data"))
    expect_true(nrow(dbGetQuery(con, "select * from sample_data")) == 0L)

    ## Only sampleData
    sd <- data.frame(fl = basename(fls), sample_name = c("A", "B"),
                     sample_index = c(1, 4))
    sampleData(m) <- DataFrame(sd)
    expect_message(dbWriteSampleData(m), "overwrite")
    res <- dbGetQuery(con, "select * from sample_data")
    expect_equal(res[, colnames(sd)], sd)

    ## sampleData and sample to spectra mapping
    m <- linkSampleData(m, with = "sampleData.fl = spectra.f")
    expect_message(dbWriteSampleData(m), "overwrite")
    res <- dbGetQuery(con, "select * from sample_data")
    expect_equal(res[, colnames(sd)], sd)
    expect_equal(dbListTables(con),
                 c("msms_spectrum", "msms_spectrum_peak_blob2",
                   "sample_data", "sample_to_msms_spectrum"))
    res <- dbGetQuery(con, "select * from sample_to_msms_spectrum")
    res <- unname(as.matrix(res))
    expect_equal(res, m@sampleDataLinks[["spectra"]])
    dbDisconnect(con)

    ## Same with a MsBackendOfflineSql
    sql_file <- tempfile()
    s <- setBackend(spectra(mse), MsBackendOfflineSql(),
                    dbname = sql_file, drv = SQLite(),
                    BPPARAM = SerialParam())
    s$f <- basename(dataOrigin(s))
    spectra(m) <- s

    dbWriteSampleData(m)
    con <- dbConnect(SQLite(), dbname = sql_file)
    expect_equal(dbListTables(con),
                 c("msms_spectrum", "msms_spectrum_peak_blob2",
                   "sample_data", "sample_to_msms_spectrum"))
    res <- dbGetQuery(con, "select * from sample_to_msms_spectrum")
    res <- unname(as.matrix(res))
    expect_equal(res, m@sampleDataLinks[["spectra"]])
    dbDisconnect(con)
})

test_that(".db_get_sample_data works", {
    res <- .db_get_sample_data(spectra(mse))
    expect_equal(res, data.frame())

    res <- .db_get_sample_data(spectra(mse_db))
    expect_equal(res[, 1:3], as.data.frame(sampleData(mse_db)))
})

test_that(".db_get_sample_spectra_link works", {
    res <- .db_get_sample_spectra_link(spectra(mse))
    expect_equal(res, matrix(ncol = 2, nrow = 0))

    res <- .db_get_sample_spectra_link(spectra(mse_db))
    expect_equal(res, matrix(ncol = 2, nrow = 0))

    spectra(mse_db)$file <- basename(dataOrigin(spectra(mse_db)))
    mse_db <- linkSampleData(
        mse_db, with = "sampleData.mzML_file = spectra.file")
    dbWriteSampleData(mse_db)

    res <- .db_get_sample_spectra_link(spectra(mse_db))
    expect_equal(res, mse_db@sampleDataLinks[["spectra"]])
})

dbDisconnect(ref_con)

#' @title A class to store experiment files
#'
#' @aliases MsExperimentFiles-class MsExperimentFiles
#'
#' @name MsExperimentFiles
#'
#' @description
#'
#' The `MsExperimentFiles` class stores files that are part of a mass
#' spectrometry experiment. The objects are created with the
#' `MsExperimentFiles()` function.
#'
#' The files encoded in a `MsExperimentFiles` instance don't need to
#' exist on the current filesystem - sometimes, these might be created
#' in anticipation of their creation. The `existMsExperimentFiles()`
#' function can be used to verify which ones currently exist: it
#' returns a list of logicals (formally an instance of
#' [IRanges::LogicalList()] of lenghts equal to the
#' `MsExperimentFiles` used as input.
#'
#' @author Laurent Gatto
#'
#' @return `MsExperimentFiles` returns an instance of `MsExperimentFiles`.
#'
#' @examples
#' fls <- MsExperimentFiles(mzmls = c("/path/to/f1.mzML", "/path/to/f2.mzML"),
#'                          mzids = "/another/path/to/id1.mzid",
#'                          fasta = "file.fas")
#' fls
#'
#' ## A new MsExperimentFiles containing mzML or mzid files
#' fls[1]
#' fls["mzids"]
#'
#' ## The actual file names
#' fls[[1]]
#' fls[[2]]
#' fls[["fasta"]]
#'
#' ## None of the files used in this example actually exist
#' existMsExperimentFiles(fls)
NULL

#' @name MsExperimentFiles-class
#'
#' @docType class
#'
#' @exportClass MsExperimentFiles
#'
#' @importClassesFrom IRanges SimpleCharacterList
#'
#' @noRd
setClass("MsExperimentFiles",
         contains = "SimpleCharacterList")

#' @importClassesFrom Spectra Spectra
#'
#' @importClassesFrom QFeatures QFeatures
#'
#' @importClassesFrom SummarizedExperiment SummarizedExperiment
#'
#' @noRd
setClassUnion("Spectra_OR_Null", c("NULL", "Spectra"))

setClassUnion("QFeatures_OR_SummarizedExperiment",
              c("SummarizedExperiment", "QFeatures"))
setClassUnion("QFeatures_OR_SummarizedExperiment_OR_Null",
              c("NULL", "QFeatures_OR_SummarizedExperiment"))


#' @export
#'
#' @rdname MsExperimentFiles
#'
#' @importFrom IRanges CharacterList
#'
#' @importMethodsFrom S4Vectors metadata metadata<-
#'
#' @importFrom methods new as validObject
#'
#' @param ... Either a named list or a set of named vectors. All
#'     elements are coerced to characters.
#'
#' @param metadata `list()` holding arbitrary R objects as
#'     annotations.
MsExperimentFiles <- function(..., metadata = list()) {
    ans <- new("MsExperimentFiles")
    ans@listData <- CharacterList(..., compress = FALSE)@listData
    if (length(metadata)) metadata(ans) <- metadata
    if (validObject(ans)) ans
}


.valid_names_non_null <- function(object) {
    if (length(object) & is.null(names(object)))
        return("List of file names must be named.")
    NULL
}

.valid_names_non_empty <- function(object) {
    if (any(names(object) == ""))
        return("Names musn't be empty")
    NULL
}

## .valid_files_exist  <- function(object) {
## }

setValidity("MsExperimentFiles",
            function(object) {
                msg <- c(.valid_names_non_null(object),
                         .valid_names_non_empty(object))
                if (is.null(msg)) TRUE
                else msg
            })


#' @rdname MsExperimentFiles
#'
#' @param object An instance of class `MsExperimentFiles`.
#'
#' @importMethodsFrom methods show
#'
#' @exportMethod show
setMethod("show", "MsExperimentFiles",
          function(object) {
              cat("MsExperimentFiles of length ", length(object), "\n")
              IRanges:::.showAtomicList(lapply(object, basename), 10)
          })

test_that(".valid_names_non_null works", {
    x <- 1:3
    names(x) <- c("a", "b", "c")
    expect_null(MsExperiment:::.valid_names_non_null(x))
    expect_null(MsExperiment:::.valid_names_non_null(list()))
    expect_match(MsExperiment:::.valid_names_non_null(1:2), "named")
})

test_that(".valid_names_non_empty works", {
    x <- 1:3
    names(x) <- c("a", "", "b")
    expect_match(MsExperiment:::.valid_names_non_empty(x), "empty")

    names(x) <- c("a", "x", "b")
    expect_null(MsExperiment:::.valid_names_non_empty(x))
})

test_that("MsExperimentFiles constructor works", {
    res <- MsExperimentFiles()
    expect_s4_class(res, "MsExperimentFiles")
    expect_true(length(res) == 0)

    res <- MsExperimentFiles(a = 4, b = 5)
    expect_true(length(res) == 2)
    expect_equal(names(res), c("a", "b"))
    expect_true(validObject(res))
})

test_that("show,MsExperimentFiles works", {
    expect_output(show(MsExperimentFiles()), "MsExperimentFiles")
})

---
title: "Managing Mass Spectrometry Experiments"
output:
    BiocStyle::html_document:
        toc_float: true
vignette: >
    %\VignetteIndexEntry{Managing Mass Spectrometry Experiments}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
    %\VignettePackage{MsExperiment}
    %\VignetteDepends{BiocStyle,rpx,Spectra,MsDataHub}
---

```{r style, echo = FALSE, results = 'asis', message=FALSE}
BiocStyle::markdown()
```

**Package**: `r BiocStyle::Biocpkg("MsExperiment")`<br />
**Authors**: `r packageDescription("MsExperiment")[["Author"]] `<br />
**Last modified:** `r file.info("MsExperiment.Rmd")$mtime`<br />
**Compiled**: `r date()`

```{r, echo = FALSE, message = FALSE}
library(MsExperiment)
library(Spectra)
library(BiocStyle)
```

# Introduction

The goal of the `r Biocpkg("MsExperiment")` package is to provide a container
for all data related to a mass spectrometry (MS) experiment. Also other
Bioconductor packages allow to represent MS experiment data (such as the
`r Biocpkg("MSnbase")` package). The `MsExperiment` however aims at being very
light-weight and flexible to accommodate all possible types of MS experiments
(proteomics, metabolomics, ...) and all types of MS data representations
(chromatographic and spectral data, quantified features etc). In addition, it
allows to bundle additional files and data, such as annotations, within the
object.

In this vignette, we will describe how to create a `MsExperiment` object and
populate it with various types of data.

```{r load_pkg, message = FALSE}
library("MsExperiment")
```

We will also use the `r Biocpkg("Spectra")` package to import MS data and thus
load it here too.

```{r load_spectra, message = FALSE}
library("Spectra")
```


# Installation

The package can be installed with the `r Biocpkg("BiocManager")` package. To
install `BiocManager` use `install.packages("BiocManager")` and, after that,
`BiocManager::install("MsExperiment")` to install `r Biocpkg("MsExperiment")`
which will install the package including all required dependencies.


# Getting data

We will use a small subset of the
[PXD022816](https://www.ebi.ac.uk/pride/archive/projects/PXD022816)
project ([Morgenstern et
al. (2020)](https://doi.org/10.1021/acs.jproteome.0c00956)). The
acquisitions correspond to a Pierce Thermo HeLa digestion standard,
diluted to 50ng/uL with 97:3 + 0.1% formic acid, and acquired on a
QExactive instrument.

Below, we use the `r Biocpkg("rpx")` package to access the project
from the PRIDE repository, and download files of interest. Note that
these will automatically be cached in the `rpx` packages' cache
directory.

```{r px, eval = TRUE}
library("rpx")
px <- PXDataset("PXD022816")
px
pxfiles(px)
```

The project provides the vendor raw files, the converted mzML files as
well as the identification mzid files. Let's download fractions 1 and
2 of the mzML files.

If you run these commands interactively and it's the first time you
use `pxget()`, you will be asked to create the `rpx` cache directory -
you can safelfy answer *yes*. The files will then be downloaded. Next
time you want to get the same files, they will be loaded automatically
from cache.

```{r fls, eval = TRUE}
(i <- grep(".+0[12].+mzML$", pxfiles(px), value = TRUE))
fls <- pxget(px, i)
fls
```

# Mass spectrometry experiment

Let's start by creating an empty `MsExperiment` object that we will
populate with different pieces of data as we proceed with the analysis
of our data.

```{r make_exp, eval = TRUE}
msexp <- MsExperiment()
msexp
```

## Experiment files

Let's now start with our MS experiment management by saving the
relevant files in a dedicated `MsExperimentFiles` object. In addition
to the mzML files, let's also assume we have the human proteomics
fasta file ready. Later, when loading the raw data into R, we will
refer directly to the files in this `MsExperimentFiles` object.

```{r make_exp_fls, eval = TRUE}
msfls <- MsExperimentFiles(mzmls = fls,
                           fasta = "homo_sapiens.fasta")
msfls
```

Let's add these files to the main experiment management object:

```{r add_exp_fls, eval = TRUE}
experimentFiles(msexp) <- msfls
msexp
```

## Experimental design

The `sampleData` slot is used to describe the overall experimental design of the
experiment. It can be used to specify the samples of the experiment and to
relate them to the files that are part of the experiment. There can be a
one-to-one link between a sample and a file, such as for example in label-free
approaches, or one-to-many, in labelled multiplexed approaches.

Here, we create a simple data frame with sample annotations that include the
original file names and the respective fractions.

```{r add_sample_data, eval = TRUE}
sampleData(msexp) <- DataFrame(
    mzmls = basename(experimentFiles(msexp)[["mzmls"]]),
    fractions = 1:2)
sampleData(msexp)
```

## Raw data

We can now create a `Spectra` object containing the raw data stored in
the mzML files. If you are not familiar with the `Spectra` object,
please refer to the [package
vignettes](https://rformassspectrometry.github.io/Spectra/articles/Spectra.html).

```{r sp, eval = TRUE}
sp <- Spectra(experimentFiles(msexp)[["mzmls"]])
sp
```

We can now add this object to the main experiment management object:

```{r add_sp, eval = TRUE}
spectra(msexp) <- sp
msexp
```

## Third party applications

Let's now assume we want to search the spectra in our mzML files
against the `homo_sapiens.fasta` file. To do so, we would like to use
a search engine such as MSGF+, that is run using the command line and
generates mzid files.

The command to run MSGF+ would look like this (see the [manual
page](https://msgfplus.github.io/msgfplus/MSGFPlus.html) for details):

```
java -jar /path/to/MSGFPlus.jar \
     -s input.mzML \
     -o output.mzid
     -d proteins.fasta \
     -t 20ppm \ ## precursor mass tolerance
     -tda 1 \   ## search decoy database
     -m 0 \     ## fragmentation method as written in the spectrum or CID if no info
     -int 1     ## Orbitrap/FTICR/Lumos
```

We can easily build such a command for each of our input file:

```{r make_cmd, eval = TRUE}
mzids <- sub("mzML", "mzid", basename(experimentFiles(msexp)[["mzmls"]]))
paste0("java -jar /path/to/MSGFPlus.jar",
       " -s ", experimentFiles(msexp)[["mzmls"]],
       " -o ", mzids,
       " -d ", experimentFiles(msexp)[["fasta"]],
       " -t 20ppm",
       " -m 0",
       " int 1")
```

Here, for the sake of time and portability, we will not actually run
MSGF+, but a simple shell script that will generate mzid files in a
temporary R directory.


```{r touch, eval = TRUE}
(output <- file.path(tempdir(), mzids))
cmd <- paste("touch", output)
cmd
```

The `cmd` variable holds the two commands to be run on the command
line that will generate the new files. We can run each of these
commands with the `system()` function.

```{r system, eval = TRUE}
sapply(cmd, system)
```

Below, we add the names of the newly created files to our experiment:

```{r add_mzids, eval = TRUE}
experimentFiles(msexp)[["mzids"]] <- mzids
experimentFiles(msexp)
msexp
```

We can also decide to store the commands that were used to generate
the mzid files in the experiment's metadata slot. Here, we use the
convention to name that metadata item `"mzmls_to_mzids"` to document
to input and output of these commands.

```{r add_cmd, eval = TRUE}
metadata(msexp)[["mzmls_to_mzids"]] <- cmd
metadata(msexp)
```

Finally, the `existMsExperimentFiles()` can be used at any time to
check which of files that are associated with an experiment actually
exist:

```{r exp_exists, eval = TRUE}
existMsExperimentFiles(msexp)
```

# Saving and reusing experiments

The `MsExperiment` object has been used to store files and data
pertaining to a mass spectrometry experiment. It is now possible to
save that object and reload it later to recover all data and metadata.
See also section *Using `MsExperiment` with `MsBackendSql`* below for an
alternative to load/restore defined MS experiments from an SQL database.

```{r save_exp, eval = TRUE}
saveRDS(msexp, "msexp.rds")
rm(list = ls())
```

```{r read_exp, eval = TRUE}
msexp <- readRDS("msexp.rds")
msexp
experimentFiles(msexp)
```

We can access the raw data as long as the mzML files that were used to
generate it still exist in their original location, which is the case
here as they were saved in the `rpx` cache directory.

```{r plot_sp, eval = TRUE}
sp <- spectra(msexp)
sp
plotSpectra(sp[1000])
```


# Linking experimental data to samples

For some experiments and data analyses an explicit link between data, data files
and respective samples is required. Such links enable an easy (and error-free)
subset or re-ordering of a whole experiment by sample and would also
simplify coloring and labeling of the data depending on the sample or of its
variables or conditions.

Below we generate an `MsExperiment` object for a simple experiment consisting of
a single sample measured in two different injections to the same LC-MS setup.

```{r make_exp2}
lmse <- MsExperiment()
sd <- DataFrame(sample_id = c("QC1", "QC2"),
                sample_name = c("QC Pool", "QC Pool"),
                injection_idx = c(1, 3))
sampleData(lmse) <- sd
```

We next add mzML files to the experiment for the sample that was measured. These
are available through Bioconductor's *MsDataHub*. We add also an additional
*annotation* file `"internal_standards.txt"` to the experiment, which could be
e.g. a file with m/z and retention times of internal standards added to the
sample (note that such files don't necessarily have to exist).

```{r add_fls2}
library(MsDataHub)
fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
         MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
basename(fls)

experimentFiles(lmse) <- MsExperimentFiles(
    mzML_files = fls,
    annotations = "internal_standards.txt")
```

Next we load the MS data from the mzML files as a `Spectra` object and add them
to the experiment (see the vignette of the `r BiocStyle::Biocpkg("Spectra")` for
details on import and representation of MS data).

```{r add_sp2}
sps <- Spectra(fls, backend = MsBackendMzR())
spectra(lmse) <- sps
lmse
```

At this stage we have thus sample annotations and MS data in our object, but no
explicit relationships between them. Without such linking between files and
samples any subsetting would only subset the `sampleData` but not any of the
potentially associated files. We next use the `linkSpectraData()` function
to establish and define such relationships. First we link the experimental files
to the samples: we want to link the first mzML file in the element called
`"mzML_file"` in the object's `experimentFiles` to the first row in `sampleData`
and the second file to the second row.

```{r link_sample_data}
lmse <- linkSampleData(lmse, with = "experimentFiles.mzML_file",
                        sampleIndex = c(1, 2), withIndex = c(1, 2))
```

To define the link we have thus to specify with which *element* within our
`MsExperiment` we want to link samples. This can be done with the parameter
`with` that takes a single `character` representing the name (*address*) of the
data element. The name is a combination of the name of the slot within the
`MsExperiment` and the name of the element (or column) within that slot
separated by a `"."`. Using `with = "experimentFiles.mzML_file"` means we want
to link samples to values within the `"mzML_file"` element of the object's
`experimentFiles` slot - in other words, we want to link samples to values in
`experimentFiles(lmse)$mzML_file`. The indices of the rows (samples) in
`sampleData` and the indices of the values in `with` to which we want to link
the samples can be defined with `sampleIndex` and `withIndex`. In the example
above we used `sampleIndex = c(1, 2)` and `withIndex = c(1, 2)`, thus, we want
to link the first row in `sampleData` to the first value in `with` and the
second row to the second value. See also the section *Linking sample data to
other experimental data* in the documentation of `MsExperiment` for more
information and details.

What happened internally by the call above is illustrated in the figure
below. The link is represented as a two-column integer `matrix` with the indices
of the linked sample in the first and the indices of the associated elements in
the second columns (this matrix is essentially a
`cbind(sampleIndex, withIndex)`).

```{r, echo = FALSE}
knitr::include_graphics("imgs/Links_01.png")
```

We next establish a second link between each sample and the *annotation*
file `"internal_standards.txt"` in `experimentFiles(lmse)$standards`:

```{r link_sample_data2}
lmse <- linkSampleData(lmse, with = "experimentFiles.annotations",
                        sampleIndex = c(1, 2), withIndex = c(1, 1))
```

The figure below illustrates again what happened internally by this call: a new
*link matrix* was added establishing the relationship between the two samples
and the one value in `experimentFiles(lmse)$annotations`.

```{r, echo = FALSE}
knitr::include_graphics("imgs/Links_02.png")
```

It is thus also possible to link different samples to the same element. We next
link the spectra in the object to the individual samples. We use for that an
alternative way to specify the link without the need to provide `sampleIndex`
and `withIndex`. Sample-to-data links can also be specified using a syntax
similar to an SQL join:
`"sampleData.<column in sampleData> = <slot>.<element in slot>"`. Links will be
thus established between elements with matching values in the specified data
fields (i.e. between rows in `sampleData` for which values in the specified
column matches values in `<slot>.<element>`). In order to use this alternative
approach to link spectra to the respective samples we have to first add the
(full) raw file name as an additional column to the object's `sampleData`. We
can now add links between spectra and samples by matching this raw file name to
the original file name from which the spectra were imported (which is available
in the `"dataOrigin"` spectra variable).

```{r link_sample_data3}
sampleData(lmse)$raw_file <- normalizePath(fls)
lmse <- linkSampleData(
    lmse, with = "sampleData.raw_file = spectra.dataOrigin")
```

The link was thus established between matching values in
`sampleData(lmse)$raw_file` and `spectra(lmse)$dataOrigin`.

```{r show_link}
sampleData(lmse)$raw_file
spectra(lmse)$dataOrigin |>
                head()
```

The figure below illustrates this link. With that last call we have thus
established links between samples and 3 different data elements in
the `MsExperiment`.

```{r, echo = FALSE}
knitr::include_graphics("imgs/Links_03.png")
```

```{r show_exp2}
lmse
```

A convenience function to quickly extract the index of a sample a spectrum is
associated with is `spectraSampleIndex()`. This function returns an `integer`
vector of length equal to the number of spectra in the object with the row in
the object's `sampleData` a spectrum is linked to, or `NA_integer_` if a
spectrum is not linked to any sample.

```{r}
#' Show the sample assignment for the first few spectra
spectraSampleIndex(lmse) |>
    head()
```

If we had also quantified *feature* values, we could also link them to the
samples. Below we create a simple, small `SummarizedExperiment` to represent
such quantified feature values and add that to our experiment. To show that
`MsExperiment` supports also links between subsets of data elements, we create a
`SummarizedExperiment` that contains values for an additional sample which is
not present in our `sampleData`. Also, we add samples in an arbitrary order.

```{r add_se, message = FALSE}
library(SummarizedExperiment)
sd <- DataFrame(sample = c("QC2", "QC1", "QC3"), idx = c(3, 1, 5))
se <- SummarizedExperiment(colData = sd, assay = cbind(1:10, 11:20, 21:30))

qdata(lmse) <- se
```

Next we link the samples in this `SummarizedExperiment` to the samples to in the
`MsExperiment` using matching values between the `"sample_id"` column in the
object's `sampleData` data frame and the column `"sample"` in the
`SummarizedExperiment`'s `colData` which is stored in the `@qdata` slot. The
naming convention to define such columns is `<slot name>.<column name>`.

```{r link_sample_data4}
sampleData(lmse)$sample_id
qdata(lmse)$sample

lmse <- linkSampleData(lmse, with = "sampleData.sample_id = qdata.sample")
lmse
```

The main advantage of all these links is that any subsetting of the experiment
by sample will keep the (linked) data consistent. To illustrate this we subset
below the experiment to the second sample.

```{r subset_exp2}
b <- lmse[2]
b
```

The subset object contains now all data elements that are linked to this
second sample. Accessing the `assay` of the `SummarizedExperiment` in
`qdata` will thus return only the quantified feature abundances for this
second sample.

```{r extract_exp2}
assay(qdata(b))
```

But what happens for data elements that are not linked to any sample? Below we
add a `data.frame` as a `metadata` to the experiment and subset the object
again.

```{r add_metadata}
metadata(lmse)$other <- data.frame(
                  sample_name = c("study_1", "POOL", "study_2"),
                  index = 1:3)
b <- lmse[2]
metadata(b)
```

By default, any element which is **not** linked to a sample is retained in the
filtered/subset object.

We next link each sample to the second row in this data frame and subset the
data again to the second sample.

```{r link_sample_data5}
lmse <- linkSampleData(lmse, with = "metadata.other",
                      sampleIndex = 1:2, withIndex = c(2, 2))
b <- lmse[2]
metadata(b)
```

Subsetting thus retained only the row in the data frame for the linked
sample. Obviously it is also possible to subset to multiple samples, in
arbitrary order. Below we re-order our experiment.

```{r show_subset_exp2}
lmse <- lmse[c(2, 1)]
sampleData(lmse)
```

The sample order is thus reversed and also all other linked elements are
re-ordered accordingly, such as `"mzML_file"` in the object's `experimentFiles`.

```{r show_expfls2_msml}
experimentFiles(lmse)$mzML_file
```

It is however important to note, that subsetting will also *duplicate* elements
that are associated with multiple samples:

```{r show_explfs2_annot}
experimentFiles(lmse)$annotations
```

Thus, while we added a single *annotation* file to the data element
`"annotations"` in `experimentFiles`, after subsetting we ended up with two
identical files. This duplication of *n:m* relationships between
samples to elements does however not affect data consistency. A sample will
always be linked to the correct value/element.


# Subset and filter `MsExperiment`

As already shown above, `MsExperiment` objects can be subset with `[` which will
subset the data by sample. Depending on whether relationships (links) between
samples and any other data within the object are present also these are
correctly subset. In addition to this general subset operation, it is possible
to individually filter the spectra data within an `MsExperiment` using the
`filterSpectra` function. This function takes any filter function supported by
`Spectra` with parameter `filter`. Parameters for this filter function can be
passed through `...`. As an example we filter below the spectra data of our
`MsExperiment` keeping only spectra with an retention time between 200 and 210
seconds.

```{r}
#' Filter the Spectra using the `filterRt` function providing also the
#' parameters for this function.
res <- filterSpectra(lmse, filterRt, rt = c(200, 210))
res
```

The resulting `MsExperiment` contains now much fewer spectra. `filterSpectra`
did only filter the spectra data, but not any of the other data slots. It did
however update and consolidate the relationships between samples and spectra
(if present) after filtering:

```{r}
#' Extract spectra of the second sample after filtering
spectra(res[2L])
```


# Using `MsExperiment` with `MsBackendSql`

The `r Biocpkg("MsBackendSql")` provides functionality to store mass
spectrometry data in a SQL database and to *represent* such data as a `Spectra`
object (through its `MsBackendSql` or `MsBackendOfflineSql` backends). Next to
the MS data, it is also possible to store sample data in such SQL databases
hence allowing to create self-contained databases for MS experiments. This
simplifies the process to load pre-defined MS experiment or to share experiment
data across analyses or with a collaborator. In this section we show how MS data
and sample information from an experiment can be stored into a SQL database and
how that data can then be loaded again as a `MsExperiment`.

Below we first define the raw data files of the experiment. In our example we
use two test files provided by Bioconductor's *MsDataHub*.

```{r}
fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
         MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
```

We can then use the `createMsBackendSqlDatabase` function from the
`r Biocpkg("MsBackendSql")` package to write the data into a SQLite
database. SQLite databases are particularly easy to create, use and also to
share, but for large experiments it is suggested to use more powerful database
engines, such as for example *MariaDB* (see also the documentation and vignette
of the `r Biocpkg("MsBackendSql")` package for parameters and settings for such
large databases).

Also, in our example we store the SQLite database into a data temporary file,
but for a real use case we would want to use a *real* file.

```{r}
library(MsBackendSql)
library(RSQLite)

#' Create the SQLite database where to store the data. For our
#' example we create the database in a temporary file.
sql_file <- tempfile()
con <- dbConnect(SQLite(), dbname = sql_file)

#' Write the MS data to the.
createMsBackendSqlDatabase(dbcon = con, fls)
```

We can now create a `Spectra` object representing the MS data in that database
using either a `MsBackendSql` or `MsBackendOfflineSql` backend.

```{r}
#' Load the database as a Spectra object
sps <- Spectra(sql_file, source = MsBackendOfflineSql(), drv = SQLite())
sps
```

Next we define sample annotations for the two files and create a `MsExperiment`
with the MS data and the related sample annotations.

```{r}
#' Define sample annotations. Ideally all experiment-relavant information
#' should be specified.
sdata <- data.frame(file_name = basename(fls),
                    sample_name = c("QC_POOL_1", "QC_POOL_3"),
                    sample_group = c("QC", "QC"))

#' To simplify subsequent linking between samples and data files we
#' define a new spectra variable with the file names of the raw data files.
sps$file_name <- basename(dataOrigin(sps))

#' Create the MsExperiment for that data
mse <- MsExperiment(sampleData = sdata, spectra = sps)
```

Finally we *link* the MS spectra to the individual samples using the file names
of the original data files.

```{r}
#' Establish the mapping between spectra and samples
mse <- linkSampleData(mse, with = "sampleData.file_name = spectra.file_name")
mse
```

With sample annotations and the mapping between samples and spectra defined, we
can next write this information to the SQL database containing already the MS
data of our experiment: the `dbWriteSampleData` function writes, for
`MsExperiment` objects that use a `Spectra` with a `MsBackendSql` (or
`MsBackendOfflineSql`) backend the available sample data to additional database
tables in the database.

```{r}
#' Write sample data (and mapping between samples and spectra) to the
#' database
dbWriteSampleData(mse)
```

The SQL database does now contain the MS data, the sample annotations and the
mapping between samples and spectra. Creation of such a database needs to be
performed only once for an experiment. The associated SQLite database file could
now be shared with a collaborator or used across different analyses without the
need to share or copy also the raw data files or sample annotations (which would
usually needed to be provided as a separate file).

To load the MS experiment data again into R, we first need to create a `Spectra`
object with a `MsBackendSql` or `MsBackendOfflineSql` backend for that SQL
database.

```{r}
#' Create a Spectra object for the MS data of that database
sps <- Spectra(sql_file, source = MsBackendOfflineSql(), drv = SQLite())
```

We then pass this `Spectra` object with parameter `spectra` to the
`MsExperiment` constructor call. The function will then read also sample data
from the database (if available) as well as the mapping between samples and
spectra to restore our MS experiment.

```{r}
#' Create the MsExperiment reading MS data and sample data from
#' the database
mse <- MsExperiment(spectra = sps)
mse
```

We have thus now the full MS experiment data available.

For smaller experiments, such SQL databases could also be created with an
alternative, eventually simpler, approach. For that we first use the
`readMsExperiment` function to create an `MsExperiment` from the file names of
the raw MS data files and a `data.frame` with the associated sample
annotations. We re-use here the variables `fls` with the data file names and
`sdata` with the sample data.

```{r}
#' Create an MsExperiment with data from two (raw) data files
#' and associated sample annotations.
mse <- readMsExperiment(spectraFiles = fls, sampleData = sdata)
mse
```

The MS data in that `MsExperiment` is represented by a `Spectra` object using
the `MsBackendMzR`:

```{r}
#' Show the Spectra from the experiment
spectra(mse)
```

We next change the backend of this `Spectra` object using the `setBackend`
function to a `MsBackendOfflineSql`. We choose again a SQLite database format
and store the data to a temporary file.

```{r}
#' Define the SQLite database file
sql_file2 <- tempfile()

#' Change the backend of the Spectra data
spectra(mse) <- setBackend(spectra(mse), MsBackendOfflineSql(),
                           dbname = sql_file2, drv = SQLite())

spectra(mse)
```

All MS data is now stored in the SQLite database and the `Spectra` object of our
experiment uses the `MsBackendOfflineSql` to represent/interface this data. To
store also the sample data to the database we can use the `dbWriteSampleData` as
before.

```{r}
#' Store also sample data to the database
dbWriteSampleData(mse)
```

In addition to create self-contained SQL databases for an experiment, this code
could also be used within an analysis workflow to serialize/de-serialize
results, always with the advantage of keeping the full experiment data in a
single place (file) in a language-agnostic format.

We can for example use basic functions from the *DBI*/*RSQLite* packages (or
simple SQL commands) to access the data in the database. Below we connect to the
database and list the available tables.

```{r}
#' Connect to the database
con <- dbConnect(SQLite(), dbname = sql_file2)

#' List the available database tables
dbListTables(con)
```

The *msms_spectrum* table contains general metadata for the individual MS
spectra, *msms_spectrum_peak_blob* the actual peak data (*m/z* and intensity
values), *sample_data* the sample annotations and *sample_to_msms_spectrum* the
link between samples and spectra. Below we retrieve the content of the
*sample_data* table.

```{r}
#' Get the content of the sample_data table
dbGetQuery(con, "select * from sample_data")

#' Close the connection to the database
dbDisconnect(con)
```

# Session information

```{r si}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsExperiment-functions.R
\name{readMsExperiment}
\alias{readMsExperiment}
\title{Import MS spectra data of an experiment}
\usage{
readMsExperiment(spectraFiles = character(), sampleData = data.frame(), ...)
}
\arguments{
\item{spectraFiles}{\code{character} with the (absolute) file names of the MS
data files that should be imported as a \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object.}

\item{sampleData}{\code{data.frame} or \code{DataFrame} with the sample annotations.
Each row is expected to contain annotations for one file (sample). The
order of the data frame's rows is expected to match the order of the
provided files (with parameter \code{spectraFiles}).}

\item{...}{additional parameters for the \code{\link[Spectra:Spectra]{Spectra::Spectra()}} call to
import the data.}
}
\value{
\code{MsExperiment}.
}
\description{
Read/import MS spectra data of an experiment from the respective (raw)
data files into an \code{\link[=MsExperiment]{MsExperiment()}} object. Files provided with the
\code{spectraFiles} parameter are imported as a \code{Spectra} object and each
file is automatically \emph{linked} to rows (samples) of a \code{sampleData}
data frame (if provided).
}
\examples{

## Define the files of the experiment to import. For our example we get
## the files from the *MsDataHub*
library(MsDataHub)
fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
         MsDataHub::X20171016_POOL_POS_3_105.134.mzML())

## Define a data frame with some sample annotations
ann <- data.frame(
    injection_index = c(1, 9),
    sample_id = c("QC_1", "QC_3"))

## Import the data
library(MsExperiment)
mse <- readMsExperiment(spectraFiles = fls, ann)
mse

## Access the spectra data
spectra(mse)

## Access the sample annotations
sampleData(mse)

## Import the data reading all MS spectra directly into memory
mse <- readMsExperiment(spectraFiles = fls, ann,
    backend = Spectra::MsBackendMemory())
mse
}
\author{
Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsExperiment-functions.R, R/MsExperiment.R
\docType{class}
\name{experimentFiles}
\alias{experimentFiles}
\alias{experimentFiles<-}
\alias{sampleData}
\alias{sampleData<-}
\alias{qdata}
\alias{qdata<-}
\alias{spectraSampleIndex}
\alias{MsExperiment}
\alias{MsExperiment-class}
\alias{show,MsExperiment-method}
\alias{length,MsExperiment-method}
\alias{spectra,MsExperiment-method}
\alias{spectra<-,MsExperiment-method}
\alias{otherData}
\alias{otherData<-}
\alias{linkSampleData}
\alias{[,MsExperiment,ANY,ANY,ANY-method}
\alias{filterSpectra,MsExperiment,function-method}
\title{Managing Mass Spectrometry Experiments}
\usage{
experimentFiles(object)

experimentFiles(object) <- value

sampleData(object)

sampleData(object) <- value

qdata(object)

qdata(object) <- value

spectraSampleIndex(x, duplicates = c("first", "keep"))

MsExperiment(
  experimentFiles = MsExperimentFiles(),
  otherData = List(),
  qdata = NULL,
  sampleData = DataFrame(),
  spectra = NULL
)

\S4method{show}{MsExperiment}(object)

\S4method{length}{MsExperiment}(x)

\S4method{spectra}{MsExperiment}(object)

\S4method{spectra}{MsExperiment}(object) <- value

otherData(object)

otherData(object) <- value

linkSampleData(
  object,
  with = character(),
  sampleIndex = seq_len(nrow(sampleData(object))),
  withIndex = integer(),
  subsetBy = 1L
)

\S4method{[}{MsExperiment,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{filterSpectra}{MsExperiment,function}(object, filter, ...)
}
\arguments{
\item{object}{An instance of class \code{MsExperiment}.}

\item{value}{An object of the appropriate class for the slot to be
populated.}

\item{x}{an \code{MsExperiment}.}

\item{duplicates}{for \code{spectraSampleIndex()}: \code{character(1)} defining the
type of result returned by \code{spectraSampleIndex()}. With
\code{duplicates = "first"} an \code{integer} vector is returned with the first
match while \code{duplicates = "keep"} returns a \code{list} of \code{integer} with
the index of all matches.}

\item{experimentFiles}{\code{\link[=MsExperimentFiles]{MsExperimentFiles()}} defining (external) files
to data or annotation.}

\item{otherData}{\code{List} with arbitrary additional (\emph{other}) information or
data.}

\item{qdata}{\code{QFeatures} or \code{SummarizedExperiment} with the quantification
data.}

\item{sampleData}{\code{DataFrame} (or \code{data.frame}) with information on
individual samples of the experiment.}

\item{spectra}{\code{\link[Spectra:Spectra]{Spectra::Spectra()}} object with the MS spectra data of the
experiment.}

\item{with}{for \code{linkSampleData()}: \code{character(1)} defining the data to which
samples should be linked. See section \emph{Linking sample data to other
experimental data} for details.}

\item{sampleIndex}{for \code{linkSampleData()}: \code{integer} with the indices of the
samples in \code{sampleData(object)} that should be linked.}

\item{withIndex}{for \code{linkSampleData()}: \code{integer} with the indices of the
elements in \code{with} to which the samples (specified by \code{sampleIndex})
should be linked to.}

\item{subsetBy}{for \code{linkSampleData()}: optional \code{integer(1)} defining the
dimension on which the subsetting will occurr on the linked data.
Defaults to \code{subsetBy = 1L} thus subsetting will happen on the first
dimension (rows or elements).}

\item{i}{for \code{[}: an \code{integer}, \code{character} or \code{logical} referring to the
indices or names (rowname of \code{sampleData}) of the samples to subset.}

\item{j}{for \code{[}: not supported.}

\item{...}{optional additional parameters. For \code{filterSpectra()}: parameters
to be passed to the filter function (parameter \code{filter}).}

\item{drop}{for \code{[}: ignored.}

\item{filter}{for \code{filterSpectra()}: any filter function supported by
\code{\link[Spectra:Spectra]{Spectra::Spectra()}} to filter the spectra object (such as \code{filterRt} or
\code{filterMsLevel}). Parameters for the filter function can be passed
through \code{...}.}
}
\value{
See help of the individual functions.
}
\description{
The \code{MsExperiment} class allows the storage and management of all
aspects related to a complete proteomics or metabolomics mass
spectrometry experiment. This includes experimantal design (i.e. a table
with samples), raw mass spectromtry data as spectra and chromatograms,
quantitative features, and identification data or any other relevant data
files.

For details, see https://rformassspectrometry.github.io/MsExperiment

This package is part of the RforMassSpectrometry initiative:
https://www.rformassspectrometry.org/
}
\section{Slots}{

\describe{
\item{\code{experimentFiles}}{An instance of class \code{MsExperimentFiles} or \code{NULL}.}

\item{\code{spectra}}{An instance of class \code{Spectra} or \code{NULL}.}

\item{\code{qdata}}{An instance of class \code{QFeatures}, \code{SummarizedExperiment} or
\code{NULL}.}

\item{\code{otherData}}{A \code{List} to store any additional data objects.}

\item{\code{sampleData}}{A \code{DataFrame} documenting the experimental design.}

\item{\code{sampleDataLinks}}{A \code{List} with link definitions between samples and
data elements. Should not be directly accessed or modified by the user.}

\item{\code{metadata}}{A \code{list} to store additional metadata.}
}}

\section{General information}{


An experiment is typically composed of several items
\itemize{
\item Description and information (covariates etc) of each sample from
the experiment. These are stored in the \code{sampleData} slot as a
\code{DataFrame}, each row describing a sample with columns containing
all relevant information on that sample.
\item Files to data or annotations. These are stored in the
\verb{@experimentFiles} slot as an instance of class \code{MsExperimentFiles}.
\item General metadata about the experiment, stored as a \code{list} in the
\verb{@metadata} slot.
\item Mass spectrometry data. Sectra and their metadata are stored as
an \verb{[Spectra::Spectra()]} object in the \code{spectra} slot. Chromatographic
data is not yet supported but will be stored as a \code{Chromatograms()}
object in the \verb{@chromatorgrams} slot.
\item Quantification data is stored as \code{QFeatures} or
\code{SummarizedExperiment} objects in the \verb{@qdata} slot and can be accessed or
replaced with the \code{qdata()} or \verb{qdata<-} functions, respectively.
\item Any additional data, be it other spectra data, or proteomics
identification data (i.e peptide-spectrum matches defined as
\code{PSM} objects) can be added as elements to the list stored in
the \code{otherData} slot.
}

The \emph{length} of a \code{MsExperiment} is defined by the number of samples (i.e.
the number of rows of the object's \code{sampleData}). A \code{MsExperiment} with two
samples will thus have a length of two, independently of the number of files
or length of raw data in the object. This also defines the subsetting of the
object using the \code{[} function which will always subset by samples. See the
section for filtering and subsetting below for more information.

\code{MsExperiment} objects can be created using the \code{MsExperiment()} function
providing the data with the parameters listed below. If the
\code{\link[Spectra:Spectra]{Spectra::Spectra()}}
object provided with the \code{spectra} param uses a \code{MsBackendSql} backend,
sample data could be retrieved from the associated SQL database (see
section \emph{Using \code{MsExperiment} with \code{MsBackendSql}} in the vignette for
details). Alternatively, it is also possible to subsequently add data and
information to an existing \code{MsExperiment}.
Finally, with the \code{\link[=readMsExperiment]{readMsExperiment()}} function it is possible to create
a \code{MsExperiment} by importing MS spectra data directly from provided data
files. See examples below or the package vignette for more information.
}

\section{Accessing data}{


Data from an \code{MsExperiment} object can be accessed with the dedicated
accessor functions:
\itemize{
\item \code{experimentFiles()}, \verb{experimentFiles<-}: gets or sets experiment files.
\item \code{length()}: get the \emph{length} of the object which represents the number of
samples availble in the object's \code{sampleData}.
\item \code{metadata()}, \verb{metadata<-}: gets or sets the object's metadata.
\item \code{sampleData()}, \verb{sampleData<-}: gets or sets the object's sample data
(i.e. a \code{DataFrame} containing sample descriptions).
\item \code{spectra()}, \verb{spectra<-}: gets or sets spectra data. \code{spectra()} returns a
\code{\link[Spectra:Spectra]{Spectra::Spectra()}} object, \verb{spectra<-} takes a \code{Spectra} data as input
and returns the updated \code{MsExperiment}.
\item \code{spectraSampleIndex()}: depending on parameter \code{duplicates} it returns
either an \code{integer} (\code{duplicates = "first"}, the default) or a \code{list}
(\code{duplicates = "keep"}) of length equal to the number of spectra within
the object with the indices of the sample(s) (in \code{sampleData()}) a
spectrum is assigned to. With \code{duplicates = "first"}, an \code{integer} with
the index is returned for each spectrum. If a spectrum was assigned to
more than one sample a warning is shown and only the first sample index
is returned for that spectrum. For \code{duplicates = "keep"}, assignments are
returned as a \code{list} of \code{integer} vectors, each element being the
index(es) of the sample(s) a spectrum is assigned to. For spectra that are
not linked to any sample an \code{NA_integer_} is returned as index for
\code{duplicates = "first"} and an empty integer (\code{integer()}) for
\code{duplicates = "keep"}.
Note that the default \code{duplicates = "first"} will work in almost all use
cases, as generally, a spectrum will be assigned to a single sample.
\item \code{qdata()}, \verb{qdata<-}: gets or sets the quantification data, which can be a
\code{QFeatures} or \code{SummarizedExperiment}.
\item \code{otherData()} , \verb{otherData<-}: gets or sets the addition data
types, stored as a \code{List} in the object's \code{otherData} slot.
}
}

\section{Linking sample data to other experimental data}{


To start with, an \code{MsExperiment} is just a loose collection of files and data
related to an experiment, no explicit links or associactions are present
between the samples and related data. Such links can however be created with
the \code{linkSampleData()} function. This function can establish links between
individual (or all) samples within the object's \code{sampleData} to individual,
or multiple, data elements or files, such as \code{Spectra} or raw data files.

The presence of such links enables a (consistent) subsetting of an
\code{MsExperiment} by samples. Thus, once the link is defined, any subsetting by
sample will also correctly subset the linked data. All other, not linked,
data elements are always retained as in the original \code{MsExperiment}.

To be able to link different elements within an \code{MsExperiment} it is also
required to \emph{identify} them with a consistent naming scheme. The naming
scheme of slots and data elements within follows an SQL-like scheme, in which
the variable (element) is identified by the name of the database table,
followed by a \code{"."} and the name of the database table column. For
\code{MsExperiment}, the naming scheme is defined as
\code{"<slot name>.<element name>"}. A column called \code{"sample_name"} within the
\code{sampleData} data frame can thus be addressed with
\code{"sampleData.sample_name"}, while \code{spectra.msLevel} would represent the
spectra variable called \code{msLevel} within the \code{Spectra} stored in the
\code{spectra} slot.

Links between sample data rows and any other data element are stored as
\code{integer} matrices within the \verb{@sampleDataLinks} slot of the object (see also
the vignette for examples and illustrations). The first column of a matrix
is always the index of the sample, and the second column the index of the
element that is linked to that sample, with one row per element.
Links can be defined/added with the \code{linkSampleData()} function which adds
a relationship between rows in \code{sampleData} to elements in any other data
within the \code{MsExperiment} that are specified with parameter \code{with}.
\code{linkSampleData()} supports two different ways to define the link:
\itemize{
\item Parameter \code{with} defines the data to which the link should be established.
To link samples to raw data files that would for example be available as a
\code{character} in an element called \code{"raw_files"} within the object's
\code{experimentFiles}, \code{with = experimentFiles.raw_files} would have to be
used. Next it is required to specify which samples should be linked with
which elements in \code{with}. This needs to be defined with the parameters
\code{sampleIndex} and \code{withIndex}, both are expected to be \code{integer} vectors
specifying which sample in \code{sampleData} should be linked to which element
in \code{with} (see examples below or vignette for examples and details).
\item As an alternative way, a link could be defined with an SQL-like syntax
that relates a column in \code{sampleData} to a column/element in the data to
which the link should be established. To link for example individual
spectra to the corresponding samples
\code{with = "sampleData.raw_file = spectra.dataOrigin"} could be used assuming
that \code{sampleData} contains a column named \code{"raw_file"} with the (full path)
of the raw data file for each sample from which the spectra were imported.
In this case both \code{sampleIndex} and \code{withIndex} can be omitted, but it is
expected/required that the columns/elements from \code{sampleData} and the data
element to which the link should be established contain matching values.
}

Note that \code{linkSampleData} will \strong{replace} a previously existing link to the
same data element.
\itemize{
\item \code{spectraSampleIndex()} is a convenience function that extracts for each
spectrum in the object's \code{spectra()} the index of the sample it is
associated with (see function's help above for more information).
}
}

\section{Subsetting and filtering}{

\itemize{
\item \code{[}: \code{MsExperiment} objects can be subset \strong{by samples} with \verb{[i]}
where \code{i} is the index or a logical defining to which samples the data
should be subset. Subsetting by sample will (correctly) subset all
linked data to the respective samples. If multiple samples are linked to
the same data element, subsetting might duplicate that data element. This
duplication of \emph{n:m} relationships between samples to elements does however
not affect data consistency (see examples below for more information).
Not linked data (slots) will be returned as they are. Subsetting in
arbitrary order is supported.
See the vignette for details and examples.
\item \code{filterSpectra()}: subsets the \code{Spectra} within an \code{MsExperiment} using a
provided filter function (parameter \code{filter}). Parameters for the filter
function can be passed with parameter \code{...}. Any of the filter functions
of a \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object can be passed with parameter \code{filter}.
Possibly present relationships between samples and spectra (\emph{links}, see
also \code{linkSampleData()}) are updated. Filtering affects only the spectra
data of the object, none of the other slots and data (e.g. \code{sampleData})
are modified.
The function returns an \code{MsExperiment} with the filtered \code{Spectra} object.
}
}

\examples{

## An empty MsExperiment object
msexp <- MsExperiment()
msexp

example(MsExperimentFiles)
experimentFiles(msexp) <- fls
msexp

## Linking samples to data elements

## Create a small experiment
library(S4Vectors)
mse <- MsExperiment()
sd <- DataFrame(sample_id = c("QC1", "QC2"),
                sample_name = c("QC Pool", "QC Pool"),
                injection_idx = c(1, 3))
sampleData(mse) <- sd

## define file names containing spectra data for the samples and
## add them, along with other arbitrary files to the experiment
fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
         MsDataHub::X20171016_POOL_POS_3_105.134.mzML())
experimentFiles(mse) <- MsExperimentFiles(
    mzML_files = fls,
    annotations = "internal_standards.txt")

## Link samples to data files: first sample to first file in "mzML_files",
## second sample to second file in "mzML_files"
mse <- linkSampleData(mse, with = "experimentFiles.mzML_files",
    sampleIndex = c(1, 2), withIndex = c(1, 2))

## Link all samples to the one file in "annotations"
mse <- linkSampleData(mse, with = "experimentFiles.annotations",
    sampleIndex = c(1, 2), withIndex = c(1, 1))
mse

## Import the spectra data and add it to the experiment
library(Spectra)
spectra(mse) <- Spectra(fls, backend = MsBackendMzR())

## Link each spectrum to the respective sample. We use the alternative
## link definition that does not require sampleIndex and withIndex but
## links elements based on matching values in the specified data elements.
## We need to add the full file name as an additional column to sampleData
## in order to allow matching this file names with the value in
## spectra(mse)$dataOrigin which contains the original file names from which
## the spectra were imported.
sampleData(mse)$raw_file <- normalizePath(fls)

## The links can be added using the short notation below
mse <- linkSampleData(mse, with = "sampleData.raw_file = spectra.dataOrigin")
mse

## With sampleData links present, any subsetting of the experiment by sample
## will ensure that all linked elements are subset accordingly
b <- mse[2]
b
sampleData(b)
experimentFiles(b)$mzML_files

## The `spectraSampleIndex()` function returns, for each spectrum, the
## index in the object's `sampleData` to which it is linked/assigned
spectraSampleIndex(mse)

## Subsetting with duplication of n:m sample to data relationships
##
## Both samples were assigned above to one "annotation" file in
## `experimentFiles`:
experimentFiles(mse[1])[["annotations"]]
experimentFiles(mse[2])[["annotations"]]

## Subsetting will always keep the relationship between samples and linked
## data elements. Subsetting will however possibly duplicate data elements
## that are shared among samples. Thus, while in the original object the
## element "annotations" has a single entry, subsetting with [1:2] will
## result in an MsExperiment with duplicated entries in "annotations"
experimentFiles(mse)[["annotations"]]
experimentFiles(mse[1:2])[["annotations"]]

## Spectra within an MsExperiment can be filtered/subset with the
## `filterSpectra` function and any of the filter functions supported
## by `Spectra` objects. Below we restrict the spectra data to spectra
## with a retention time between 200 and 210 seconds.
res <- filterSpectra(mse, filterRt, rt = c(200, 210))
res

## The object contains now much less spectra. The retention times for these
rtime(spectra(res))

## Relationship between samples and spectra was preserved by the filtering
a <- res[1L]
spectra(a)
}
\author{
Laurent Gatto, Johannes Rainer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsExperiment-db.R
\name{dbWriteSampleData}
\alias{dbWriteSampleData}
\title{Write sample annotations to a MsBackendSql SQL database}
\usage{
dbWriteSampleData(x)
}
\arguments{
\item{x}{\code{MsExperiment} from which sample annotations should be written
to the database.}
}
\description{
For \code{MsExperiment} objects with their MS data represented by a \code{Spectra}
object that use a \code{MsBackendSql} backend, its sample annotations can be
written to the backend's SQL database with the \code{dbWriteSampleData()}
function.
The content of the object's \verb{[sampleData()]} (as well as eventually present
\emph{linking} between samples and spectra) will be stored in two separate
database tables \emph{sample_data} and \emph{sample_to_msms_spectrum} in the same
database.

This requires that the MS data of the experiment is \emph{represented} by a
\code{MsBackendSql} backend (see help on the \code{createMsBackendSqlDatabase} or the
\emph{MsBackendSql} package vignette for more information on how to create or
use such SQL databases).
}
\examples{

library(MsExperiment)

## Create a MsBackendSql database from two mzML files.
## Connect first to an empty SQLite database (for the example we create
## a database in a temporary file).
library(RSQLite)
sqlite_db <- tempfile()
con <- dbConnect(SQLite(), sqlite_db)

## Define the files from which we import the data. Test files provided from
## the *MsDataHub*
fls <- c(MsDataHub::X20171016_POOL_POS_1_105.134.mzML(),
    MsDataHub::X20171016_POOL_POS_3_105.134.mzML())

## Create a MsBackendSql database containing the full MS data
library(MsBackendSql)
createMsBackendSqlDatabase(con, fls)

## Note: alternatively it would be possible to first import the MS data
## to a `Spectra` object and then change the backend to a `MsBackendSql`
## using the `setBackend` function.

## Load this data as a `Spectra` object (using a `MsBackendOfflineSql`
## backend)
library(Spectra)
sps <- Spectra(sqlite_db, source = MsBackendOfflineSql(),
    drv = SQLite())
sps

## Define sample annotations for the two data files. Adding one column
## `"file"` that contains the file name of the data files.
df <- data.frame(sample = c("QC1", "QC2"), file = basename(fls))

## Add a spectra variable `"file"` to the `Spectra` object with
## the raw data files' file names to simplify the linking between
## samples and spectra performed later.
sps$file <- basename(dataOrigin(sps))

## Create a MsExperiment with the spectra and sample data.
mse <- MsExperiment(spectra = sps, sampleData = df)

## Establish the link (mapping) between samples and spectra
## using the column `"file"` in the `sampleData` and the spectra
## variable `"file"`.
mse <- linkSampleData(mse, with = "sampleData.file = spectra.file")
mse

## Write sample data (and the sample to spectra mapping) to the
## *MsBackendSql* database.
dbWriteSampleData(mse)

## List the tables in the database
dbListTables(con)

## Sample data was thus stored to the database.
dbGetQuery(con, "select * from sample_data;")
}
\author{
Johannes Rainer, Laurent Gatto
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsExperimentFiles.R, R/existMsExperimentFiles.R
\name{MsExperimentFiles}
\alias{MsExperimentFiles}
\alias{MsExperimentFiles-class}
\alias{show,MsExperimentFiles-method}
\alias{existMsExperimentFiles}
\title{A class to store experiment files}
\usage{
MsExperimentFiles(..., metadata = list())

\S4method{show}{MsExperimentFiles}(object)

existMsExperimentFiles(object)
}
\arguments{
\item{...}{Either a named list or a set of named vectors. All
elements are coerced to characters.}

\item{metadata}{\code{list()} holding arbitrary R objects as
annotations.}

\item{object}{The \code{existMsExperimentFiles()} function works with
either an instance of \code{MsExperimentFiles} or \code{MsExperiment}.}
}
\value{
\code{MsExperimentFiles} returns an instance of \code{MsExperimentFiles}.
}
\description{
The \code{MsExperimentFiles} class stores files that are part of a mass
spectrometry experiment. The objects are created with the
\code{MsExperimentFiles()} function.

The files encoded in a \code{MsExperimentFiles} instance don't need to
exist on the current filesystem - sometimes, these might be created
in anticipation of their creation. The \code{existMsExperimentFiles()}
function can be used to verify which ones currently exist: it
returns a list of logicals (formally an instance of
\code{\link[IRanges:AtomicList-class]{IRanges::LogicalList()}} of lenghts equal to the
\code{MsExperimentFiles} used as input.
}
\examples{
fls <- MsExperimentFiles(mzmls = c("/path/to/f1.mzML", "/path/to/f2.mzML"),
                         mzids = "/another/path/to/id1.mzid",
                         fasta = "file.fas")
fls

## A new MsExperimentFiles containing mzML or mzid files
fls[1]
fls["mzids"]

## The actual file names
fls[[1]]
fls[[2]]
fls[["fasta"]]

## None of the files used in this example actually exist
existMsExperimentFiles(fls)
}
\author{
Laurent Gatto
}
